/******************************************************************************
Last Updata 2007/09/27 by Pascu Moca
No documentation updata necessary.

******************************************************************************/


#include "../include/cut_off.h"

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void get_m_star(int iteration_index,
		int max_state_no, 
		int new_block_state_no,
		int * m_star){

/***********************************************************************
input variables = three
      iteration_index  - the index of the iteration in the NRG cycle. 
      int max_state_no - maximum number of block states allowed to represent
                         the block (kept states). It is defined by the user 
                         in the input.dat file a priory of the run. 
      int new_block_state_no - requested number of block states to represent the
                         new block which is formed from the block of the
                         previous iteration step and an extra spin. 

output variables = one
      int * m_star     - the iteration index at which truncation begins. 
                         m_star remains fixed for the rest of the NRG steps. 

return = void

What it does:
     This routine computes the m_star that is the iteration_index at which 
     truncation of the states begins (new_block_state_no > max_state_no), i.e.,   
     we start to descard states. For iterations larger then m_star 
     the states, transformation_matrices, spectral operators have to be saved. 
     m_star remains fixed for the rest of the NRG steps. 

how to call:
***********************************************************************/

  if(new_block_state_no > max_state_no){
    if((*m_star)==-1){
      (*m_star) = iteration_index;
    }//if
  }//if

}//get_m_star

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void set_states_elimination_tags(int iteration_index,
				 int max_state_no, 

				 int state_no, 
				 state *states_array,

				 double *new_block_energies_array, 
				 double *sorted_new_block_energies_array,
				 int & kept_state_no){

/***********************************************************************
input variables = five
    iteration_index           - index of the current iteration in the NRG run. 
    max_state_no              - maximum number of states that are kept. 
                                (parameter readed from the input.dat file)
    state_no                  - the number of the states that has to be labeled. 
    new_block_energies_array        - array with energies of the given states.
    sorted_new_block_energies_array - array with energies of the states sorted 
                                in ascending order. 

output variables = one
    states_array              - array with the states with the elimination_tag 
                                property set to either 'K' or 'D'.

return = void

What it does:
    This routine sets the labels 'K' for kept and 'D' for discarded on the 
    states. First the energies are arranged in ascending order and if the 
    no of states is larger than the max_state_no, those states with larger energy 
    has to be discarded. They will be assigned as 'D' while all the kept
    states will have the elimination_tag set to 'K'. 
how to call:
***********************************************************************/

  if(state_no<=max_state_no){
    kept_state_no = state_no;
    return; 
  }//if

  if(state_no>max_state_no){

    kept_state_no = max_state_no; 
    while((sorted_new_block_energies_array[kept_state_no] - sorted_new_block_energies_array[kept_state_no - 1] < degeneracy_threshhold) 
	  && (kept_state_no > 0)) 
      kept_state_no--;    

    
    if(kept_state_no == 0){
      cout<<"The number of the kept states is zero.\nError in routine set_states_elimination_tags().\nEXITING..."<<endl;
      exit(-1); 
    }//if

    for(int state_index=0; state_index<state_no; state_index++){
      
      if(new_block_energies_array[state_index] > sorted_new_block_energies_array[kept_state_no - 1]){
	states_array[state_index].set_elimination_tag('D'); 
      }//if 
    }//for state_index    
  }//if
  
  if (iteration_index==iteration_no){
    for(int state_index=0; state_index<state_no; state_index++){
      states_array[state_index].set_elimination_tag('D'); 
    }//for state_index
  }
  

  return; 
}//set_states_elimination_tags

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void transfer_states_and_sectors(int symmetry_no, 
				 symmetry * symmetry_types,
				 
				 int max_state_no, 
				 int kept_state_no,
		       
				 int  new_block_state_no, 
				 state * new_block_states_array,
 
				 int  new_block_sector_no,
				 sector_properties * new_block_sectors_properties_array, 
				 
				 int & block_state_no, 
				 state * & block_states_array,

				 int & block_sector_no,
				 sector_properties * & block_sectors_properties_array,
				 
				 double *new_block_energies_array, 
				 double *sorted_new_block_energies_array){

/***********************************************************************

input variables = nine
     symmetry_no      - the number of symmetries in the problem. 
     symmetry_types   - the types of symmetries invoved. 

     max_state_no     - the maximul number of states kept at one iteration 
                        (readed from the input.dat file)

     new_block_state_no     - the number of new states that has to be cut-off.
     new_block_states_array - the new states that has to converted to the current 
                        states.

     new_block_sector_no    - the number of new sectors.
     new_block_sectors_properties_array - arrays with the properties of the new 
                                    sectors.

     new_block_energies_array - array with the values of the eigenenergies of 
                          the Hamiltonian at this iteration
     sorted_new_block_energies_array - array with the sorted values of the 
                                 eigenenergies of the Hamiltonian

output variables = four
     block_state_no     - the number of the block_states
     block_states_array - the actually current states obtained from the new_block_states_array; 

     block_sector_no - the 'new' number of the current sectors base
     block_sectors_properties_array - 'new' properties for the current sectors. 

return = void

What it does:
     Once the new_block_states_array is computed and we have to move to the next iteration 
     we have to transfer the new_block_states_array to the block_states_array. 
     This routine transfer the new_block_states_array to the block_states_array 
     such that the iteration cycle can be restarted. 

how to call:
***********************************************************************/
  
  
    if(kept_state_no == 0){
      cout<<"The number of the kept states is zero.\nError in routine transfer_states().\nEXITING..."<<endl;
      exit(-1); 
    }//if

    if (kept_state_no > max_state_no){
       cout<<"The number of the kept states is larger than the max_state_no.\nError in routine transfer_states().\nEXITING..."<<endl;
      exit(-1); 
    }//if


  if (kept_state_no > new_block_state_no){
       cout<<"The number of the kept states is larger than the new_block_state_no.\nError in routine transfer_states().\nEXITING..."<<endl;
      exit(-1); 
    }//if


  /*if we did not discarded any state the kept_state_no is the same with the new_block_state_no
    we can copy the new_block_states into the block_states directly. */


  
  if (kept_state_no == new_block_state_no){

    block_state_no = new_block_state_no;


    if(block_states_array) {
      delete [] block_states_array;
    }
    
    block_states_array = NULL;
    
    block_states_array = new state [block_state_no];




    if (block_original_sector_array){
      delete [] block_original_sector_array;
    }


    block_original_sector_array = new int [block_state_no]; 


    if (block_original_sector_index_array){
      delete [] block_original_sector_index_array;
    }
    
    block_original_sector_index_array = new int [block_state_no]; 
    


    for(int state_index = 0; state_index < block_state_no; state_index++){      
      block_states_array[state_index] = new_block_states_array[state_index];
      block_original_sector_array[state_index] = block_states_array[state_index].get_sector();
      block_original_sector_index_array[state_index] = block_states_array[state_index].get_sector_index();
    }//for state_index
    
    for(int sector_index=0; sector_index<block_sector_no; sector_index++){
      block_sectors_properties_array[sector_index].representation_index.clear();
    }//for sector_index


    if (block_sectors_properties_array){
      delete [] block_sectors_properties_array;
    }//if

    block_sectors_properties_array = NULL;


    block_sector_no = new_block_sector_no;    
    block_sectors_properties_array = new sector_properties [block_sector_no];
    
    for(int sector_index = 0; sector_index < block_sector_no; sector_index++){
      block_sectors_properties_array[sector_index].representation_index.allocate(symmetry_no);
      block_sectors_properties_array[sector_index].representation_index = new_block_sectors_properties_array[sector_index].representation_index;
      block_sectors_properties_array[sector_index].dimension = new_block_sectors_properties_array[sector_index].dimension;
      block_sectors_properties_array[sector_index].first_state = new_block_sectors_properties_array[sector_index].first_state; 
    }//for sector_index
  










  }//if kept_state_no==new_block_state_no
  


  if (kept_state_no < new_block_state_no){


    block_state_no = kept_state_no;

    if (block_states_array){
      delete [] block_states_array;
    }

    block_states_array = NULL;

    block_states_array = new state [block_state_no];


    if (block_original_sector_array){
      delete [] block_original_sector_array;
    }

    
    block_original_sector_array = new int [block_state_no]; 
    
    
    if (block_original_sector_index_array){
      delete [] block_original_sector_index_array;
    }
    
    block_original_sector_index_array = new int [block_state_no]; 
    

    int new_block_state_index(0);

    int kept_state_index(0);
     
    while(kept_state_index < kept_state_no){

      if(new_block_states_array[new_block_state_index].get_elimination_tag() == 'K'){
	block_states_array[kept_state_index] = new_block_states_array[new_block_state_index];
	block_original_sector_array[kept_state_index] =  new_block_states_array[new_block_state_index].get_sector();
	block_original_sector_index_array[kept_state_index] =  new_block_states_array[new_block_state_index].get_sector_index();
	new_block_state_index++;
	kept_state_index++;
      }//if 
      else{
	  new_block_state_index++;
      }//else
    }//while      


    for(int sector_index=0; sector_index<block_sector_no; sector_index++){
      block_sectors_properties_array[sector_index].representation_index.clear();
    }//for sector_index
    
    delete [] block_sectors_properties_array;
    block_sectors_properties_array = NULL;
    

    int kept_sector_no=get_sectors_no(block_state_no, block_states_array);
    block_sector_no = kept_sector_no;

    set_sectors_representation_index(symmetry_no,
				     symmetry_types,
				     block_state_no, 
				     block_states_array, 
				     block_sector_no, 
				     block_sectors_properties_array); 


  }//if kept_state_no<new_block_state_no

  return;




}//transfer_states_and_sectors

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void transfer_quantum_operator(int symmetry_no, 
			       symmetry *symmetry_types,

			       int max_state_no, 

			       int new_block_state_no, 
			       state *  new_block_states_array, 

			       int  new_block_sector_no,
			       sector_properties *  new_block_sectors_properties_array, 
				 
			       int  block_state_no, 
			       state* block_states_array,

			       int  block_sector_no,
			       sector_properties * &block_sectors_properties_array,
				 
			       quantum_operator & one_operator,
			       quantum_operator & new_one_operator, 

			       double *new_block_energies_array, 
			       double *sorted_new_block_energies_array
			       ){

/***********************************************************************
input variables = fourteen
     symmetry_no    - the number of symmetries in the problem. 
     symmetry_types - the types of symmetries invoved. 

     max_state_no     - the maximul number of states kept at one iteration 
                        (readed from the input.dat file)

     new_block_state_no     - the number of new states that has to be cut-off.
     new_block_states_array - the new states that has to converted to the 
                        current states.

     new_block_sector_no                - the number of new sectors.
     new_block_sectors_properties_array - arrays with the properties of the 
                                    new sectors.

     block_state_no     - the number of the block_states
     block_states_array - the actually current states obtained from 
                            the new_block_states_array; 

     block_sector_no                - the  number of the current sectors 
     block_sectors_properties_array -  properties for the current sectors. 

     new_one_operator - is the operator in the new basis that has to be 
                        transfer to the one_operator. 

     new_block_energies_array - array with the values of the eigenenergies of the 
                          Hamiltonian at this iteration
     sorted_new_block_energies_array - array with the sorted values of the 
                                 eigenenergies of the Hamiltonian


output variables = one
     one_operator - in a copy of the new_one_operator into the one_operator

return = void

What it does:
     This routine transfer the new_one_operator to the one_operator. The output 
     is a single operator only, therefore has to be used multiple time, once
     for each operator. 
     This routine is used to transfer the Hamiltonian, hopping_operators as 
     well as the spectral operators. 

How to call:
***********************************************************************/

  if(new_block_state_no <= max_state_no){

    one_operator.clear_sectors_matrix(); 
    one_operator.allocate_sectors_matrix(block_sector_no);

    for(int sector_row = 0; sector_row < block_sector_no; sector_row++){
      for(int sector_column = 0; sector_column < block_sector_no; sector_column++){
	
	one_operator.get_sector_matrix()[sector_row][sector_column].set_size(block_sectors_properties_array[sector_row].dimension,
									     block_sectors_properties_array[sector_column].dimension);

	if(new_one_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){	  
	  
	  one_operator.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	  one_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, block_sectors_properties_array[sector_row].representation_index);
	  one_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, block_sectors_properties_array[sector_column].representation_index);
	  one_operator.get_sector_matrix()[sector_row][sector_column].allocate_matrix();
	  one_operator.get_sector_matrix()[sector_row][sector_column].get_matrix() = new_one_operator.get_sector_matrix()[sector_row][sector_column].get_matrix();
	  
	}//if
      }//sector_column
    }//sector_row
  }//if new_block_state_no

  
  if(new_block_state_no>max_state_no){

    one_operator.clear_sectors_matrix(); 
    one_operator.allocate_sectors_matrix(block_sector_no);


    for(int sector_row = 0; sector_row < block_sector_no; sector_row++){
      for(int sector_column = 0; sector_column < block_sector_no; sector_column++){

	one_operator.get_sector_matrix()[sector_row][sector_column].set_size(block_sectors_properties_array[sector_row].dimension,
									     block_sectors_properties_array[sector_column].dimension);
	if(grand_selection_rule(symmetry_no, symmetry_types, block_sectors_properties_array[sector_column].representation_index,
				one_operator.get_representation_index(),block_sectors_properties_array[sector_row].representation_index)) {


	  one_operator.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	  one_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, block_sectors_properties_array[sector_row].representation_index);
	  one_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, block_sectors_properties_array[sector_column].representation_index);
	  one_operator.get_sector_matrix()[sector_row][sector_column].allocate_matrix();

	  int new_block_sector_row(0);
	  int new_block_sector_column(0);

	  for(new_block_sector_row = 0; new_block_sector_row < new_block_sector_no; new_block_sector_row++)
	    if(new_block_sectors_properties_array[new_block_sector_row].representation_index == block_sectors_properties_array[sector_row].representation_index)
	      break;

	  for(new_block_sector_column = 0; new_block_sector_column < new_block_sector_no; new_block_sector_column++)
	    if(new_block_sectors_properties_array[new_block_sector_column].representation_index == block_sectors_properties_array[sector_column].representation_index)
	      break;

	  int first_state_row(new_block_sectors_properties_array[new_block_sector_row].first_state - 1);
	  int first_state_column(new_block_sectors_properties_array[new_block_sector_column].first_state - 1);
	  int height(one_operator.get_sector_matrix()[sector_row][sector_column].get_height());
	  int width(one_operator.get_sector_matrix()[sector_row][sector_column].get_width());
	  int new_block_height(new_one_operator.get_sector_matrix()[new_block_sector_row][new_block_sector_column].get_height());
	  int new_block_width(new_one_operator.get_sector_matrix()[new_block_sector_row][new_block_sector_column].get_width());

	  for(int i = 0; i < height; i++){
	    for(int j = 0; j < width; j++){
	      
	      int state_row = first_state_row;
	      int state_column = first_state_column;

	      int new_block_i(-1);
	      int new_block_j(-1);

	      for(int k = 0; k <= i; k++) 
		while(new_block_i < (new_block_height - 1)){
		  ++new_block_i;
		  ++state_row;
		  if(new_block_energies_array[state_row] <= sorted_new_block_energies_array[block_state_no -1])
		    break;
		}//while

	      for(int k = 0; k <= j; k++)
		while(new_block_j < (new_block_width - 1)){
		  ++new_block_j;
		  ++state_column;
		  if(new_block_energies_array[state_column] <= sorted_new_block_energies_array[block_state_no -1])
		    break;
		}//while

	      double matrix_element = new_one_operator.get_sector_matrix()[new_block_sector_row][new_block_sector_column].get_matrix()[new_block_i][new_block_j];
	      one_operator.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i, j, matrix_element);
	    }//j
	  }//i
     
	}//if grand_selection_rule
	
      }//sector_column
    }//sector_row
 
  }//if new_block_state_no> max_state_no

  new_one_operator.clear();

}//transfer_quantum_operator


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void clear_new_block_states_and_sectors(state * & new_block_states_array,
				  int new_block_sector_no, 
				  sector_properties * & new_block_sectors_properties_array, 
				  double * & new_block_energies_array,
				  double * & sorted_new_block_energies_array,
				  unsigned long * & new_block_energies_index,
				  sector * & transformation_sectors){

/***********************************************************************
input variables = five
     ???

output variables = one
     ???

return = void

What it does:
     This routine is used to clear some variables allocated dinamically 
     that are used in the NRG iteration. At each iteration these 
     have different sizes therefore has to be clear and allocated at each iteration.
     All the parameters are input parameters and the allocated quantities 
     are all deleted and the corresponding pointer are pointed to NULL.

How to call:
***********************************************************************/
  if(new_block_states_array){
    delete []  new_block_states_array; 
  }//if
  new_block_states_array = NULL;

  if(new_block_sectors_properties_array){
    for(int i=0; i<new_block_sector_no; i++){
      new_block_sectors_properties_array[i].representation_index.clear();
    }//i
    delete [] new_block_sectors_properties_array;
  }//if
  new_block_sectors_properties_array = NULL ; 
  
  if(new_block_energies_array){
    delete [] new_block_energies_array;
  }//if
  new_block_energies_array = NULL;
  
  if(sorted_new_block_energies_array){
    delete [] sorted_new_block_energies_array; 
  }//if
  sorted_new_block_energies_array = NULL; 
  
  if(new_block_energies_index){
    delete [] new_block_energies_index; 
  }//if
  new_block_energies_index = NULL; 
  
  if(transformation_sectors){
    for(int i =0; i<new_block_sector_no; i++){
      transformation_sectors[i].clear();
    }//for    
    delete [] transformation_sectors;
  }
  
  transformation_sectors=NULL; 


}//clear_new_block_state_and_sectors

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
