/*******************************************************************************
Last Updata 2007/09/25 Pascu Moca

No documentation updata necessary

*******************************************************************************/

#include "../include/diagonalization.h"

void block_hamiltonian_diagonalization(int sector_no, 
				       sector_properties *sectors_properties_array, 
				       quantum_operator & block_hamiltonian, 
				       sector * transformation_sectors, 
				       double * new_block_energies_array){
  
/*******************************************************************************

input variables = five
     int sector_no - the number of sectors in which the Block_Hamiltonian is organized.
     sector_properties *sectors_properties_array - properties of the sectors. 
     Each one has a dimension, degeneracy, first_state and representation_index. 
     quantum_operator & block_hamiltonian - 
     sector * transformation_sectors -  
     double * new_block_energies_array - 

output variables = two
     transformation_sectors - is an array of sectors such that the corresponding
     matrix of a given sector is the unitary transformation of the given 
     sector of the Block_Hamiltonian to the diagonal basis. 

     new_block_energies_array - represents the eigen_energies of the Block_Hamiltonian. This 
     vector is not ordered. Later has to be ordered. 

return = void 

What it does:
     This routine diagonalize the Block_Hamiltonian operator separately in each sector 
     and returns the unitary matrix for each sector, as well as all the 
     eigenenergies of the Block_Hamiltonian. The routine uses the LAPACK library for 
     diagonalization. 
     

how to call:
     
*******************************************************************************/


  int eigen_energy_index = 0;
  
  for(int sector_label = 0; sector_label < sector_no; sector_label++){

    /************************************************************************
     * Setting parameters of the given sector that will be diagonalized by  *
     * routine dsyev_.                                                      * 
     * n is the size of the matrix, a[n,n] is the matrix to be diagonalized.*
     * dsyev_ returns a[n,n] that contains the eigenfunctions in column form*
     * and eigenenergy_a[n] a vector that contains the eigenenergies.       *
     * new_block_energies_array[max number of state ] a vecor that contains all   *
     * eigenenergies of all sectors. Energy eigenvalues are not sorted yet. *
     * transformation_sector contains the eigenfunctions is row form,       *
     * therefore, the transpose of a[n,n] is taken.                         *
     ************************************************************************/
        
    int sector_size = sectors_properties_array[sector_label].dimension;
    
    transformation_sectors[sector_label].set_size(sector_size, sector_size);    

    transformation_sectors[sector_label].allocate_matrix();
    
    int n = (int)sector_size;
    
    double ** a = NULL;
    
    a = new double * [n];
    a[0] = new double [n * n];
    for (int i = 1; i < n; i++)
      a[i] = a[0] + i * n;
    
    double * work = NULL;
    double * eigenenergy_a = NULL;
    
    work = new double [66 * n];
    eigenenergy_a = new double [n];
   
    /* Setting parameters and work space for dsyev_ */  
    char jobz = 'V';
    char uplo = 'U'; 
    
    int lda = n;
    int lwork = 66 * n;
    int info;
   
    /* What is the reason to take the transpose ????? */ 
    for(int i = 0; i < n; i++){
      for(int j = 0; j < n; j++){
	a[i][j] = (double)block_hamiltonian.get_sector_matrix()[sector_label][sector_label].get_matrix()[j][i];
      }//j
    }//i

    /* If matrix dimension is one no need for diagonalization */    
    if(n == 1){
      eigenenergy_a[0] = (double)block_hamiltonian.get_sector_matrix()[sector_label][sector_label].get_matrix()[0][0]; ;
      transformation_sectors[sector_label].set_matrix_element(0, 0, 1);
    }//if
    
    if(n != 1){
      dsyev_( & jobz,
	      & uplo,
	      & n,
	      * a,
	      & lda,
	      eigenenergy_a,
	      work,
	      & lwork,
	      & info );
      
      assert (info ==0); 

      for(int i = 0; i < n; i++){
	for(int j = 0; j < n; j++){
	    transformation_sectors[sector_label].set_matrix_element(i, j, a[j][i]);	 
	}//for j
	  	  
    }//i
    }//if
 
    for(int i = 0; i < n; i++){
      new_block_energies_array[eigen_energy_index] = eigenenergy_a[i];
      eigen_energy_index++;
    }//for i
      
    delete [] a[0];
    a[0] = NULL;
    delete [] a;
    a = NULL;
    delete [] work;
    work = NULL;
    delete [] eigenenergy_a;
    eigenenergy_a = NULL;
  }//for sector_label
}//block_hamiltonian_diagonalization

