/****************************************************************************

Last Update : 2007/09/25 Pascu Moca 

No documentation update necessary


*****************************************************************************/


#include "../include/dmatrix.h"

/*----------------------------------------------------------------------------*/
/*---------------CONSTRUCTORS AND DESTRUCTORS---------------------------------*/
/*----------------------------------------------------------------------------*/

dmatrix::dmatrix(){

/*****************************************************************
class dmatrix
      description of variables:
           no_of_rows - this is the no of rows
           no_of_columns - this is the number of columns
           matrix_ptr - a double pointer to double array with the matrix elements.

Implicit constructor for the dmatrix class.

input variables = none

output variables  = none

return = none

What it does:
     set the 'no_of_rows' and 'no_of_columns' to zero and the 'matrix_ptr' is set to NULL

how to call:
     dmatrix();
****************************************************************/

  no_of_rows = 0; 
  no_of_columns = 0; 
  matrix_ptr = NULL;     
}//dmatrix


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

dmatrix::dmatrix(int height, int width){

/****************************************************************************
Implicit constructor for the dmatrix class.

input variables = two
        int height - 'no_of_rows' is set to this value
        int width  - 'no_of_columns' is set to this value
output variables  = none

return = none

What it does:
     set the 'no_of_rows' to the first input variable
     and 'no_of_columns' to the second input variable  and the 'matrix_ptr'
     allocates an double array of size [no_of_rows][no_of_columns]
     the indexing of the matrix_ptr is matrix_ptr[0..no_of_rows-1][0..no_of_columns-1]
     all the 'matrix_ptr' elements are set to zero.
how to call:
     dmatrix(height, width);
****************************************************************************/


  no_of_rows  = height; 
  no_of_columns = width;

  matrix_ptr = new double * [no_of_rows];
  if (!matrix_ptr){
    cout << "allocation failure 1 in dmatrix::dmatrix(int, int)\nEXITING..." << endl; 
    exit(1);
  }//if
  
  matrix_ptr[0]= new double[no_of_rows * no_of_columns];
  if (!matrix_ptr[0]){
    cout << "allocation failure 2 in dmatrix::dmatrix(int, int)\nEXITING..." << endl; 
    exit(1);
  }//if
	
  /* Now I set the pointers of the other rows */
  for(int i = 1; i < no_of_rows; i++){
    matrix_ptr[i] = matrix_ptr[i-1] + no_of_columns;
  }//for i 
  for(int i = 0; i < height; i++){
    for(int j = 0; j < width; j++){
      matrix_ptr[i][j] = 0;
    }//j
  }//i
  
}//dmatrix();

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

dmatrix:: ~dmatrix(){

/****************************************************************************
Implicit destructor for the dmatrix class.

input variables = none

output variables  = none

return = none

What it does:
     distroys the dmatrix class, by setting the 'no_of_columns' and
     'no_of_rows' to zero and 'matrix_ptr' to NULL.

how to call:
     ~dmatrix();
****************************************************************************/

  if(matrix_ptr){
    no_of_rows  = 0; 
    no_of_columns = 0;
    
    delete [] matrix_ptr[0];
    matrix_ptr[0] = NULL;
    delete [] matrix_ptr;
    matrix_ptr = NULL; 
  }//if

}//~dmatrix.

/*----------------------------------------------------------------------------*/
/*-----------------------METHODS----------------------------------------------*/
/*----------------------------------------------------------------------------*/

    
void dmatrix::allocate(int height, int width){

/****************************************************************************
input variables = two
        int height - 'no_of_rows' is set to this value
        int width  - 'no_of_columns' is set to this value
output variables  = none

return = void

What it does:
     set the 'no_of_rows' to the first input variable
     and 'no_of_columns' to the second input variable  and the 'matrix_ptr'
     allocates an double array of size [no_of_rows][no_of_columns]
     the indexing of the matrix_ptr is matrix_ptr[0..no_of_rows-1][0..no_of_columns-1]
     and all the 'matrix_ptr' elements are set to zero.
how to call:
     allocate(height, width);
****************************************************************************/

  if(!matrix_ptr) {
    no_of_rows  = height; 
    no_of_columns = width;
    /* allocate pointers to rows */
    matrix_ptr = new double * [no_of_rows];
    if (!matrix_ptr){
      cout << "allocation failure 1 in dmatrix:allocate(int, int)\nEXITING..." << endl; 
      exit(1);
    }//if

     
    /* allocate rows and set pointers to them */
    matrix_ptr[0] = new double[no_of_rows * no_of_columns];
    if (!matrix_ptr[0]){
      cout << "allocation failure 2 in dmatrix::allocate(int, int)\nEXITING..." << endl; 
      exit(1);
    }//if
       
    /* Now I set the pointers of the other rows */
    for(int i = 1; i < no_of_rows; i++){
      matrix_ptr[i]=matrix_ptr[i-1] + no_of_columns;
    }//for i
    for(int i = 0; i < height; i++)
      for(int j = 0; j < width; j++)
	matrix_ptr[i][j] = 0;
  }//if matrix_ptr 
  else{
    cerr << "The matrix is already allocated.\nEXITING..." << endl; 
    exit(1);
  }//else  
}//allocate


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void dmatrix::clear(void){

/****************************************************************************
input variables = none

output variables  = none

return = void

What it does:
     set to zero the  'no_of_rows' and 'no_of_columns' and the 'matrix_ptr'
     is set to NULL.
how to call:
     clear();
****************************************************************************/

  if(matrix_ptr) {  
    /* if allocated, otherwise it is clear */ 
    no_of_rows  = 0; 
    no_of_columns = 0;
    /* remove arrays */
    delete [] (matrix_ptr[0]);
    matrix_ptr[0] = NULL; 
    delete [] matrix_ptr;
    matrix_ptr = NULL; 
  }//if matrix_ptr
  else{
    cerr << "The matrix is not yet allocated.Error in dmatrix::clear()\nEXITING..." << endl; 
    exit(1);
  }//else
}//clear

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
    
int  dmatrix::get_no_of_columns(){

/****************************************************************************
input variables = none

output variables  = none

return = an integer that represent the 'no_of_columns' variable.

What it does:
     returns the value of the 'no_of_columns' variable
how to call:
     get_no_of_columns();
****************************************************************************/

  return(no_of_columns);
}//no_of_columns

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int  dmatrix::get_no_of_rows(void){

/****************************************************************************
input variables = none

output variables  = none

return = an integer that represent the 'no_of_rows' variable.

What it does:
     returns the value of the 'no_of_rows' variable
how to call:
     get_no_of_rows();
****************************************************************************/

  return(no_of_rows);
}//no_of_rows; 

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

double ** dmatrix::get_matrix_ptr(void) {

/****************************************************************************
input variables = none

output variables  = none

return = an double pointer to a double variable that represents the 'matrix_ptr' variable.

What it does:
     returns the 'matrix_ptr' pointer.
how to call:
     get_matrix_ptr();
****************************************************************************/

  return(matrix_ptr);
}//prt

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void dmatrix::set_matrix_element(int i, int j, double x){

/****************************************************************************
input variables = three
        int i - is the corresponding row in the matrix table that will be modified.
        int j - is the corresponding column in the matrix table that will be modified
        double x - is the new value of the matrix_ptr[i][j] element.

output variables  = none

return = void

What it does:
      set the 'matrix_ptr[i][j]' to x.
how to call:
     get_matrix_ptr(i,j, x);
****************************************************************************/

  if((matrix_ptr) && (i < no_of_rows) && (j < no_of_columns)) {
    matrix_ptr[i][j] = x;
  }//if 
  else {
    cerr << "Illegal action. Cannot set the matrix element.\nError in dmatrix::set_matrix_element(int, int, double)\nEXITING..." << endl; 
    exit(1);
  }//else
}//set_matrix_element
    
/*----------------------------------------------------------------------------*/
/*------------------------OPERATORS-------------------------------------------*/
/*----------------------------------------------------------------------------*/

double * dmatrix::operator [] (int i){

/****************************************************************************
input variables = one
        int i - is the corresponding row in the matrix table.

output variables  = none

return = a pointer to a double that represents the address of the
         'i-th' row in the 'matrix_ptr' array.

What it does:
      returns a pointer to the i-th row in the 'matrix_ptr' variable.
how to call:
      dmatrix::operator[](int)
****************************************************************************/

  return(matrix_ptr[i]);
}//[]
    
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


dmatrix & dmatrix::operator = (dmatrix & mmatrix){

/****************************************************************************
input variables = one
        dmatrix &  mmatrix - is a new dmatrix variable that in copied in the current one.

output variables  = none

return = a reference to the new dmatrix variable. Now the 'matrix_ptr' contains
         the mmatrix.matrix_ptr elements.

What it does:
      copy the mmatrix variable in the present one.
how to call:
      dmatrix::operator =(dmatrix &)
****************************************************************************/

  if(no_of_columns != mmatrix.no_of_columns || no_of_rows != mmatrix.no_of_rows) {
    cerr << "The operation cannot be performed. The sizes of the matrices do not match.\nError in dmatrix::operator = (dmatrix& )\nEXITING..."
	 << endl;
    exit(1);
  }//if no_of_columns
  for(int i = 0; i < no_of_rows; i++)
    for(int j = 0; j < no_of_columns; j++)
      matrix_ptr[i][j] = mmatrix.matrix_ptr[i][j]; 
  return (*this);
}//operator =

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

dmatrix & dmatrix::operator *= (const dmatrix & mmatrix){

/****************************************************************************
input variables = one
        dmatrix &  mmatrix - is a new dmatrix variable that multiplies to the right the current one.

output variables  = none

return = a reference to the new dmatrix variable, that is the product of the current one
         with the mmatrix. The product is done in the form of matrix multiplication
         for the 'matrix_ptr' elements of the current and 'mmatrix' variable.

What it does:
      multiplies the current variable with the 'mmatrix'  in the form of matrix product.
how to call:
      dmatrix::operator *(dmatrix &)
****************************************************************************/

  if(no_of_columns != mmatrix.no_of_rows) {

    cerr << "The matrix multiplication cannot be performed. The sizes of the matrices do not match.\nError in dmatrix::operator *=(const dmatrix & )\nEXITING..."
	 << endl;
    exit(1);
  } 
  dmatrix result(no_of_rows, mmatrix.no_of_columns);
  for(int i = 0; i < no_of_rows; i++)
    for(int j = 0; j < mmatrix.no_of_columns; j++)
      for(int k = 0; k < no_of_columns; k++)
	result.matrix_ptr[i][j] += matrix_ptr[i][k] * mmatrix.matrix_ptr[k][j];
  (*this).clear();
  (*this).allocate(result.no_of_rows, result.no_of_columns);
  (*this) = result;
  return (*this);
}//operator *=

/*----------------------------------------------------------------------------*/
/*-------------------------FRIEND FUNCTIONS-----------------------------------*/
/*----------------------------------------------------------------------------*/


    /* transformation from the LHS */
dmatrix  operator * (const dmatrix & lmatrix, dmatrix & rmatrix){

/****************************************************************************
input variables = two
        dmatrix &  lmatrix - left matrix in the product
        dmatrix &  rmatrix - right matrix in the product

output variables  = none

return = a dmatrix variable that is the matrix product of the lmatrice.matrix_ptr matrix
        with the rmatrix.matrix_ptr

What it does:
      multiplies two dmatrix varibles lmatrix*rmatrix and returns an new variable.
how to call:
      new_dmatrix = lmatrix*rmatrix.
****************************************************************************/

  if(lmatrix.no_of_columns != rmatrix.no_of_rows){
    cerr << "The matrix multiplication cannot be performed. The sizes of the matrices do not match.\nError \
    in dmatrix friend operator *(const dmatrix &, dmatrix &)\nEXITING..."
	 << endl;
    exit(1);
  }//if

  dmatrix result(lmatrix.no_of_rows, rmatrix.no_of_columns);
  for(int i = 0; i < lmatrix.no_of_rows; i++)
    for(int j = 0; j < rmatrix.no_of_columns; j++)
      for(int k = 0; k < lmatrix.no_of_columns; k++)
	result.matrix_ptr[i][j] += lmatrix.matrix_ptr[i][k] * rmatrix.matrix_ptr[k][j];
  return (result);
}//*

/*----------------------------------------------------------------------------*/

