/****************************************************************************

Last Update : 2007/09/26 Pascu Moca 

Documentation update necessary

*****************************************************************************/


#include "../include/dmnrg.h"

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

 void write_transformation_sectors(int m_star,

				  long iteration_index, 

				  int state_no,
				  int sector_no, 
				  sector_properties *sectors_properties_array,

				  sector * transformation_sectors
				  ){

/*******************************************************************************
input variables:  two
    m_star                   - the iteration at which the writing process begins. 
    iteration_index          - the index of the current iteration. This variable
                               is written in the file name. 
    state_no                 - the number of the states. 
    sector_no                - the number of sectors. 
    sectors_properties_array - array with properties of the sectors. 
    transformation_sectors   - array with sectors that diagonalize the 
                               Hamiltonian these are obtained by using the 
                               hamiltonian_diagonalization(...) routine. 

output variables: none 

return: void

What it does:
    This routine writes the transformation sectors, that diagonalize the Hamiltonian
    into files separately for each iteration. m_star represents the iteration when 
    the process starts, such that the new states created is larger than the maximum
    amount of states that are kept during the iteration procedure. 

How to call:
    ???
*******************************************************************************/

 
    transformation_swap_file_name( transformations_folder_name, 
				   iteration_index,
				   transformation_file_name
				   );   
    transformation_file_out.open(transformation_file_name, ios::out); 

    if (!transformation_file_out.good()){
      error_message("writing transformation_file"); 
    }//if

    
    transformation_file_out<<"new_block_sector_no="<<sector_no<<endl;
   
    for(int sector_label = 0; sector_label < sector_no; sector_label++){
      transformation_file_out<<"sector="<<sector_label<<endl;
      long sector_dimension = sectors_properties_array[sector_label].dimension; 

      transformation_file_out<<"sector_dimension="<<sector_dimension<<endl; 
      
      for (int i=0; i<sector_dimension; i++){
	for(int j=0; j<sector_dimension; j++){
	  transformation_file_out<<setprecision(double_precision)<<transformation_sectors[sector_label].get_matrix()[j][i]<<"\t"; 
	}//j
	transformation_file_out<<endl; 
      }//i
    }//for sector_label
    transformation_file_out.close();
    transformation_file_out.clear();
   
  return; 
}//write_transformation_sectors;



/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void write_transformation_sectors_binary(int m_star,

					 long iteration_index, 
					 
					 int state_no,
					 int sector_no, 
					 sector_properties *sectors_properties_array,
					 
					 sector * transformation_sectors
					 ){

/*******************************************************************************
input variables:  two
    m_star                   - the iteration at which the writing process begins. 
    iteration_index          - the index of the current iteration. This variable
                               is written in the file name. 
    state_no                 - the number of the states. 
    sector_no                - the number of sectors. 
    sectors_properties_array - array with properties of the sectors. 
    transformation_sectors   - array with sectors that diagonalize the 
                               Hamiltonian these are obtained by using the 
                               hamiltonian_diagonalization(...) routine. 

output variables: none 

return: void

What it does:
    This routine writes the transformation sectors, that diagonalize the Hamiltonian
    into files separately for each iteration. m_star represents the iteration when 
    the process starts, such that the new states created is larger than the maximum
    amount of states that are kept during the iteration procedure. 

How to call:
    ???
*******************************************************************************/

 
    transformation_binary_swap_file_name( transformations_folder_name, 
					  iteration_index,
					  transformation_file_name
					  );   

    transformation_file_out.open(transformation_file_name, ios::out|ios::binary); 

    if (!transformation_file_out.good()){
      error_message("writing transformation_file"); 
    }//if

    
    transformation_file_out.write((const char *) &sector_no, sizeof(int));
   
    for(int sector_label = 0; sector_label < sector_no; sector_label++){
      transformation_file_out.write((const char *) &sector_label, sizeof(int));

      int sector_dimension = sectors_properties_array[sector_label].dimension; 

      transformation_file_out.write((const char *) &sector_dimension, sizeof(int)); 
      
      for (int i=0; i<sector_dimension; i++){
	for(int j=0; j<sector_dimension; j++){
	  transformation_file_out.write((const char *) &transformation_sectors[sector_label].get_matrix()[j][i], sizeof(double)); 
	}//j
      }//i
    }//for sector_label

    transformation_file_out.close();
    transformation_file_out.clear();
   
  return; 
}//write_transformation_sectors_binary;




/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void write_states(int m_star,

		  long iteration_index, 

		  int new_block_state_no,
		  state *new_block_states_array,
		  double *new_block_energies_array,

		  int block_state_no, 
		  state *block_states_array, 
		  int *block_original_sectors_array, 
		  int *block_original_sectors_index_array
		  ){

/*******************************************************************************
input variables:  five
    m_star               - the iteration at which the writing process begins. 
    iteration_index      - the index of the current iteration. This variable is 
                           written in the file name. 
    new_block_state_no         - the number of the states. 
    new_block_states_array     - array with the states that are written to the file. 
    current state_no     - the number of the current states. 
    block_states_array - array with the current states. 

output variables: none

return: void

What it does:
    This routine writes the states into the file. At each iteration a seperate
    file is written. The name of the file contains the iteration_index variable. 
    Each file has two sections. One the new_block_states_array is written together 
    with properties of each state such as the sector to which the state belongs, 
    the representation_index, the parent_state_index, local_state_index, 
    and the energy of the given state. 

    The other section corresponds the block_states, with the same quantities except
    that the energies are not written because these are not available for the current states. 

How to call:

*******************************************************************************/

    
    states_swap_file_name( states_folder_name, 
			   iteration_index,
			   states_file_name
			   );  
 
    states_file_out.open(states_file_name, ios::out); 

    if (!states_file_out.good()){
      error_message("writing states_file"); 
    }//if


    
    states_file_out<<"new_block_state_no="<<new_block_state_no<<endl;
    states_file_out<<"# sector\tsector_index\trepresentation_index\tparent_state_index\tlocal_state_index\tsgn\telimination_tag\t\t energy"<<endl; 
    for(int state_index = 0; state_index < new_block_state_no; state_index++){

      states_file_out  <<  new_block_states_array[state_index].get_sector() 
		       << "\t\t" << new_block_states_array[state_index].get_sector_index()
		       << "\t\t" << new_block_states_array[state_index].get_representation_index() 
		       << "\t\t\t\t" << new_block_states_array[state_index].get_parent_state_index() 
		       << "\t\t\t" << new_block_states_array[state_index].get_local_state_index()
		       << "\t\t" << new_block_states_array[state_index].get_sgn()
		       << "\t\t" << new_block_states_array[state_index].get_elimination_tag() 
		       << "\t\t" << new_block_energies_array[state_index]<<endl;
    }//for state_index
    
    states_file_out<<"block_state_no="<<block_state_no<<endl;
 states_file_out<<"# sector\tsector_index\trepresentation_index\tparent_state_index\tlocal_state_index\tsgn\telimination_tag\t\toriginal_sector\t\toriginal_sector_index"<<endl; 
    for(int state_index = 0; state_index < block_state_no; state_index++){

      states_file_out  <<  block_states_array[state_index].get_sector() 
		       << "\t\t" << block_states_array[state_index].get_sector_index()
		       << "\t\t" << block_states_array[state_index].get_representation_index() 
		       << "\t\t\t\t" << block_states_array[state_index].get_parent_state_index() 
		       << "\t\t\t" << block_states_array[state_index].get_local_state_index()
		       << "\t\t" << block_states_array[state_index].get_sgn()
		       <<"\t\t"<<block_states_array[state_index].get_elimination_tag()
		       <<"\t\t\t"<<block_original_sectors_array[state_index]
	               <<"\t\t\t"<<block_original_sectors_index_array[state_index]<<endl;
	
    }//for state_index
    states_file_out.close();
    states_file_out.clear();
 

  return; 
}//write_states


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void write_sectors(int m_star,

		   long iteration_index, 

		   int new_block_state_no,
		   int new_block_sector_no,
		   sector_properties *new_block_sectors_properties_array,

		  int block_sector_no, 
		  sector_properties *block_sectors_properties_array
		  ){

/*************************************************************************
input variables:  seven
     m_star                           - the iteration at which the writing 
                                        process begins. 
     iteration_index                  - the index of the current iteration. 
                                        This variable is written in the 
                                        file name. 
     new_block_state_no                     - the number of the states. 
     new_block_sector_no                    - the number of new sectors that are 
                                        written into the file. 
     new_block_sectors_properties_array     - array with the properties of the 
                                        new sectors
     current sector_no                - the number of the current sectors.
     block_sectors_properties_array - array with the current properties 
                                        of the sectors. 

output variables: none 

return: void

What it does:
    This routine writes the sectors into the file. At each iteration a seperate
    file is written. The name of the file contains the iteration_index variable. 
    Each file has two sections. One is for the new_block_sectors_properties_array 
    where all the properties of the new sectors are written. The other
    section is for the block_sectors_properties array. 

How to call:
*************************************************************************/
    
    sectors_swap_file_name( sectors_folder_name, 
			   iteration_index,
			   sectors_file_name
			    );  
     
    sectors_file_out.open(sectors_file_name, ios::out); 

    if (!sectors_file_out.good()){
      error_message("writing sectors_file"); 
    }//if

    
    sectors_file_out<<"new_block_sector_no="<<new_block_sector_no<<endl;
    sectors_file_out<<endl; 
    sectors_file_out<<"#representation_index\tdimension\tfirst_state\tdegeneracy"<<endl;
    for(int sector_index = 0; sector_index < new_block_sector_no; sector_index++){
      
      sectors_file_out  <<"\t"  << new_block_sectors_properties_array[sector_index].representation_index
			<< "\t\t" << new_block_sectors_properties_array[sector_index].dimension
			<< "\t\t" << new_block_sectors_properties_array[sector_index].first_state
			<< "\t\t" << new_block_sectors_properties_array[sector_index].degeneracy<<endl;
			

    }//for sector_index
    
    sectors_file_out<<endl; 

    sectors_file_out<<"block_sector_no="<<block_sector_no<<endl;
    sectors_file_out<<endl; 

    for(int sector_index = 0; sector_index < block_sector_no; sector_index++){
      
      sectors_file_out  << "\t" << block_sectors_properties_array[sector_index].representation_index
			<< "\t\t" << block_sectors_properties_array[sector_index].dimension
			<< "\t\t" << block_sectors_properties_array[sector_index].first_state
			<< "\t\t" << block_sectors_properties_array[sector_index].degeneracy<<endl; 
			
    }//for sector_index
    


    sectors_file_out.close();
    sectors_file_out.clear();

  
  return; 
}//write_sectors


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void write_spectral_operator(int m_star,
			     long iteration_index, 

			     int new_block_state_no,
			     int new_block_sector_no,

			     quantum_operator & spectral_operator
			     ){

/*************************************************************************
input variables:  five
    m_star            - the iteration at which the writing process begins. 
    iteration_index   - the index of the current iteration. This variable 
                        is written in the file name. 
    new_block_state_no      - the number of the states. 
    new_block_sector_no     - the number of new sectors that are written into 
                        the file. 
    spectral_operator - the operator that is written to the file. 

output variables: none

return: void

What it does:
    This routine writes one spectral operator into the file. At each iteration a seperate
    file is written. The name of the file contains the name of the spectral operator 
    and the iteration_index. 
    The routine has to be called separately for each spectral operator. 

How to call:
*************************************************************************/
    
        spectral_operator_swap_file_name( spectral_operators_folder_name, 
					  iteration_index,
					  spectral_operator_file_name,
					  spectral_operator
				      );  
        
    spectral_operator_file_out.open(spectral_operator_file_name, ios::out); 

    if (!spectral_operator_file_out.good()){
      error_message("writing spectral_operator_file"); 
    }//if

    //count the number of active sectors:

    int active_sector_no(0);
      for(int sector_row=0; sector_row<new_block_sector_no; sector_row++){
	for(int sector_column=0; sector_column<new_block_sector_no; sector_column++){
	  if(spectral_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){
	    active_sector_no++; 
	  }//if
	}//sector_column
      }//sector_row

    spectral_operator_file_out<<"active_sector_no="<<active_sector_no<<endl;
    for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++)
      for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++)
	
      if(spectral_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){

	spectral_operator_file_out<<"sector_row="<<sector_row<<endl;
	spectral_operator_file_out<<"sector_column="<<sector_column<<endl;
	
	
	int row_size =spectral_operator.get_sector_matrix()[sector_row][sector_column].get_height();
	int column_size = spectral_operator.get_sector_matrix()[sector_row][sector_column].get_width(); 

	spectral_operator_file_out<<"row_dimension="<<row_size<<endl; 
	spectral_operator_file_out<<"column_dimension="<<column_size<<endl; 
	
	
	for (int i=0; i<row_size; i++){
	  for(int j=0;j<column_size; j++){
	    spectral_operator_file_out<<setprecision(double_precision)<<spectral_operator.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j]<<"\t"; 
	  }//j
	  spectral_operator_file_out<<endl;
	}//i
				
      }//if one_operator.get_sector_matrix();

    spectral_operator_file_out.close();
    spectral_operator_file_out.clear();
    
  return; 
}//write_spectral_operator


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void write_spectral_operator_binary(int m_star,
				    long iteration_index, 
				    
				    int new_block_state_no,
				    int new_block_sector_no,
				    
				    quantum_operator & spectral_operator
				    ){

/*************************************************************************
input variables:  five
    m_star            - the iteration at which the writing process begins. 
    iteration_index   - the index of the current iteration. This variable 
                        is written in the file name. 
    new_block_state_no      - the number of the states. 
    new_block_sector_no     - the number of new sectors that are written into 
                        the file. 
    spectral_operator - the operator that is written to the file. 

output variables: none

return: void

What it does:
    This routine writes one spectral operator into the file. At each iteration a seperate
    file is written. The name of the file contains the name of the spectral operator 
    and the iteration_index. 
    The routine has to be called separately for each spectral operator. This routine
    writes the spectral operator into a binary file for space saving on the disk. 
    In this way the space is reduced up to 1/4 from the space needed when usual 
    txt files are used. The time spent on input/output is also reduced. 

How to call:
*************************************************************************/
    
    spectral_operator_binary_swap_file_name( spectral_operators_folder_name, 
					     iteration_index,
					     spectral_operator_file_name,
					     spectral_operator
					     );  
    
    spectral_operator_file_out.open(spectral_operator_file_name, ios::out| ios::binary); 
    
    if (!spectral_operator_file_out.good()){
      error_message("writing spectral_operator_file"); 
    }//if

    //count the number of active sectors:

    int active_sector_no(0);
      for(int sector_row=0; sector_row<new_block_sector_no; sector_row++){
	for(int sector_column=0; sector_column<new_block_sector_no; sector_column++){
	  if(spectral_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){
	    active_sector_no++; 
	  }//if
	}//sector_column
      }//sector_row

    spectral_operator_file_out.write((const char *) &active_sector_no, sizeof(int));

    for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++)
      for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++)
	
      if(spectral_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){

	spectral_operator_file_out.write((const char *) &sector_row, sizeof(int));
	spectral_operator_file_out.write((const char *) &sector_column, sizeof(int));;
	
	
	int row_size =spectral_operator.get_sector_matrix()[sector_row][sector_column].get_height();
	int column_size = spectral_operator.get_sector_matrix()[sector_row][sector_column].get_width(); 

	spectral_operator_file_out.write((const char *) &row_size, sizeof(int)); 
	spectral_operator_file_out.write((const char *) &column_size, sizeof(int)); 
	
	
	for (int i=0; i<row_size; i++){
	  for(int j=0;j<column_size; j++){
	    spectral_operator_file_out.write((const char *) &spectral_operator.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j], sizeof(double)); 
	  }//j
	}//i
	
      }//if one_operator.get_sector_matrix();

    spectral_operator_file_out.close();
    spectral_operator_file_out.clear();
    
  return; 
}//write_spectral_operator_binary





/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void write_static_operator(int m_star,
			     long iteration_index, 

			     int new_block_state_no,
			     int new_block_sector_no,

			     quantum_operator & static_operator
			     ){

/*************************************************************************
input variables:  five
    m_star            - the iteration at which the writing process begins. 
    iteration_index   - the index of the current iteration. This variable 
                        is written in the file name. 
    new_block_state_no      - the number of the states. 
    new_block_sector_no     - the number of new sectors that are written into 
                        the file. 
    static_operator - the operator that is written to the file. 

output variables: none

return: void

What it does:
    This routine writes one static operator into the file. At each iteration a seperate
    file is written. The name of the file contains the name of the static operator 
    and the iteration_index. 
    The routine has to be called separately for each static operator. 

How to call:
*************************************************************************/

    
        static_operator_swap_file_name( static_operators_folder_name, 
					  iteration_index,
					  static_operator_file_name,
					  static_operator
				      );  
        
    static_operator_file_out.open(static_operator_file_name, ios::out); 

    if (!static_operator_file_out.good()){
      error_message("writing static_operator_file"); 
    }//if

    //count the number of active sectors:

    int active_sector_no(0);
      for(int sector_row=0; sector_row<new_block_sector_no; sector_row++){
	for(int sector_column=0; sector_column<new_block_sector_no; sector_column++){
	  if(static_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){
	    active_sector_no++; 
	  }//if
	}//sector_column
      }//sector_row

    static_operator_file_out<<"active_sector_no="<<active_sector_no<<endl;
    for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++)
      for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++)
	
      if(static_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){

	static_operator_file_out<<"sector_row="<<sector_row<<endl;
	static_operator_file_out<<"sector_column="<<sector_column<<endl;
	
	
	int row_size =static_operator.get_sector_matrix()[sector_row][sector_column].get_height();
	int column_size = static_operator.get_sector_matrix()[sector_row][sector_column].get_width(); 

	static_operator_file_out<<"row_dimension="<<row_size<<endl; 
	static_operator_file_out<<"column_dimension="<<column_size<<endl; 
	
	
	for (int i=0; i<row_size; i++){
	  for(int j=0;j<column_size; j++){
	    static_operator_file_out<<setprecision(double_precision)<<static_operator.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j]<<"\t"; 
	  }//j
	  static_operator_file_out<<endl;
	}//i
				
      }//if one_operator.get_sector_matrix();

    static_operator_file_out.close();
    static_operator_file_out.clear();
    
  return; 
}//write_static_operator


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void write_static_operator_binary(int m_star,
				    long iteration_index, 
				    
				    int new_block_state_no,
				    int new_block_sector_no,
				    
				    quantum_operator & static_operator
				    ){

/*************************************************************************
input variables:  five
    m_star            - the iteration at which the writing process begins. 
    iteration_index   - the index of the current iteration. This variable 
                        is written in the file name. 
    new_block_state_no      - the number of the states. 
    new_block_sector_no     - the number of new sectors that are written into 
                        the file. 
    static_operator - the operator that is written to the file. 

output variables: none

return: void

What it does:
    This routine writes one static operator into the file. At each iteration a seperate
    file is written. The name of the file contains the name of the static operator 
    and the iteration_index. 
    The routine has to be called separately for each static operator. This routine
    writes the static operator into a binary file for space saving on the disk. 
    In this way the space is reduced up to 1/4 from the space needed when usual 
    txt files are used. The time spent on input/output is also reduced. 

How to call:
*************************************************************************/

    
    static_operator_binary_swap_file_name( static_operators_folder_name, 
					     iteration_index,
					     static_operator_file_name,
					     static_operator
					     );  
    
    static_operator_file_out.open(static_operator_file_name, ios::out| ios::binary); 
    
    if (!static_operator_file_out.good()){
      error_message("writing static_operator_file"); 
    }//if

    //count the number of active sectors:

    int active_sector_no(0);
      for(int sector_row=0; sector_row<new_block_sector_no; sector_row++){
	for(int sector_column=0; sector_column<new_block_sector_no; sector_column++){
	  if(static_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){
	    active_sector_no++; 
	  }//if
	}//sector_column
      }//sector_row

    static_operator_file_out.write((const char *) &active_sector_no, sizeof(int));

    for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++)
      for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++)
	
      if(static_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){

	static_operator_file_out.write((const char *) &sector_row, sizeof(int));
	static_operator_file_out.write((const char *) &sector_column, sizeof(int));;
	
	
	int row_size =static_operator.get_sector_matrix()[sector_row][sector_column].get_height();
	int column_size = static_operator.get_sector_matrix()[sector_row][sector_column].get_width(); 

	static_operator_file_out.write((const char *) &row_size, sizeof(int)); 
	static_operator_file_out.write((const char *) &column_size, sizeof(int)); 
	
	
	for (int i=0; i<row_size; i++){
	  for(int j=0;j<column_size; j++){
	    static_operator_file_out.write((const char *) &static_operator.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j], sizeof(double)); 
	  }//j
	}//i
	
      }//if one_operator.get_sector_matrix();

    static_operator_file_out.close();
    static_operator_file_out.clear();
    
  return; 
}//write_static_operator_binary





/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void write_m_star(int m_star,
		  char *m_star_file_name){

/*****************************************************************************
input variables:  two
    int m_star             - ??? 
    char *m_star_file_name - 

output variables: one

return: void 

What it does:
    This routine writes the m_star into the file in the data folder

How to call:
*****************************************************************************/

  m_star_swap_file_name(swap_folder_name,
			m_star_file_name); 

  ofstream m_star_file(m_star_file_name, ios::out); 
  
  m_star_file<<"m_star = "<<m_star<<endl;

  m_star_file.close();
  m_star_file.clear();
  return; 
}


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


double get_discarded_partition_function(int iteration_index,
					int symmetry_no, 
					symmetry * symmetry_types, 
					
					int new_block_sector_no, 
					sector_properties * new_block_sectors_properties_array, 
					
					int new_block_state_no, 
					state *new_block_states_array, 
					
					double *new_block_energies_array,
					double *sorted_new_block_energies_array,
					unsigned long *new_block_energies_index,
					
					double temperature 
    ){

/******************************************************************************
input variables:  ten
    iteration_index - the current iteration for which the partition function is computed
    symmetry_no     - is the number of symmetries in the problem. 
    symmetry_types  - represents the symmetries of the problem involved. 
 

    new_block_sector_no               - the number of new sectors . 
    new_block_sector_properties_array -  properties of the sectors. 

    new_block_state_no     - the number of the states. 
    new_block_states_array - array with the states used for the partition function calculation.

    new_block_energies_array        - array with the eigenvalues of the Hamiltonian at the present iteration. 
    sorted_new_block_energies_array - array with sorted eigenvalues of the Hamiltonian at the present iteration
    new_block_energies_index        - array which indexes the new_block_energies_array. 

output variables: none

return: double that is ???

What it does:
    This routine computes the partition function contribution from one iteration. 
    In the main_fnrg all these contributions from separate iterations are added 
    and the total partition function is finded.

How to call:
******************************************************************************/

  if (temperature==0){
    error_message("The function get_partion_function(...) cannot be called for zero temperature\nExiting..");
  }//if
  
  
  double rescaled_temperature = temperature*pow (lambda, 0.5*((double)(iteration_index))); 
  /* note that temperature = bare_temperature *2/ (1+ Lambda^-1)/sqrt(Lambda) */

  double Z_discarded_states =0.0; 

  /* compute the total degeneracy of the local states */
  double local_degeneracy = get_local_degeneracy(symmetry_no, 
						 symmetry_types, 
						 
						 new_block_sector_no, 
						 new_block_sectors_properties_array
						 ); 

  
  
  double energy_minimum = sorted_new_block_energies_array[0]; 
  /* this is actually 0, but we keep it for clarity */

  int state_sector; 
  double state_degeneracy; 

  /* we compute the contribution of the discarded states */
  for (int state_index=0; state_index<new_block_state_no; state_index++){

      /* compute the degeneracy of a given state */
    state_sector = new_block_states_array[state_index].get_sector();
    state_degeneracy = (double)new_block_sectors_properties_array[state_sector].degeneracy;

    /* add its contribution to Z */
    if(new_block_states_array[state_index].get_elimination_tag()=='D'){
      Z_discarded_states  +=( exp(-(new_block_energies_array[state_index]-energy_minimum)/rescaled_temperature))*
	pow(local_degeneracy, (double)(iteration_no-iteration_index))*state_degeneracy;
    }
   
  }//state_index

  return Z_discarded_states ;
  
}//get_partition_function


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void construct_density_matrix(int symmetry_no,
			      symmetry * symmetry_types,

			      int new_block_sector_no,
			      sector_properties * new_block_sectors_properties_array,

			      int new_block_state_no,
			      state * new_block_states_array,

			      double *new_block_energies_array,
			      double *sorted_new_block_energies_array,
			      unsigned long *new_block_energies_index,


			      quantum_operator & new_block_hamiltonian,
			      quantum_operator & new_block_density_matrix,

			      double temperature,
			      double partition_function
			      ){

/*************************************************************************
input variables:  eleven
     symmtry_no - is the number of symmetries in the problem. 
     symmetry_types - represents the symmetries of the problem involved. 

     new_block_sector_no - the number of new sectors that are written into the file. 
     new_block_sector_properties_array -  properties of the sectors. 

     new_block_state_no - the number of the states. 
     new_block_states_array - array with the states used for constructing the density matrix.

     new_block_energies_array - array with the eigenvalues of the Hamiltonian at the last iteration. 
     sorted_new_block_energies_array - array with sorted eigenvalues of the Hamiltonian at the last iteration. 
     new_block_energies_index - array which indexes the new_block_energies_array. 
     new_block_hamiltonian - Hamiltonian matrix at the last iteration 

     temperature - the temperature (this parameter is readed from the input.dat file)

output variables: one
     new_block_density_matrix - the density matrix constructed at the last iteration. 

return: void

What it does:
     This routine computes the density matrix at the last iteration. The new_block_density_matrix
     variable is organized as a quantum_operator as well with the same representation 
     index as the Hamiltonian. 

How to call:
*************************************************************************/


  double energy_min = sorted_new_block_energies_array[0];
  double energy_max = sorted_new_block_energies_array[new_block_state_no - 1];

  
  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
    if(new_block_hamiltonian.get_representation_index(symmetry_index) != 0){
      error_message("checking the quantum numbers of the Hamiltonian.\nError in routine create_new_block_hamiltonian\nExiting...\n");
    }//if
  }//for symmetry_index


  new_block_density_matrix.activate(symmetry_no); 
  new_block_density_matrix.set_name("new_block_density_matrix"); 
  new_block_density_matrix.allocate_sectors_matrix(new_block_sector_no); 

  //we set the representation indices of the new_block_density_matrix to be zero's. 

  for(int i = 0; i < symmetry_no; i++){
    new_block_density_matrix.set_representation_index(i,0); 
  }//i

  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
    if(new_block_density_matrix.get_representation_index(symmetry_index) != 0){
      error_message("checking the quantum numbers of the new_block_density_matrix.\nError in routine construct_density_matrix\nExiting...\n");
    }//if
  }//for symmetry_index
  
  
  
  for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++){
    for(int sector_column = 0; sector_column <new_block_sector_no; sector_column++){
            
      new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
										 new_block_sectors_properties_array[sector_column].dimension);

      /* we activate the diagonal part of the density matrix */
      if(grand_selection_rule(symmetry_no, 
			      symmetry_types, 
			      new_block_sectors_properties_array[sector_column].representation_index,
			      new_block_density_matrix.get_representation_index(), 
			      new_block_sectors_properties_array[sector_row].representation_index)){


	new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
	new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
	new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].allocate_matrix();

	int height = new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].get_height();
	int width = new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].get_width();

	//now we work within the sector to set up the matrix elements. 

	for(int i=0; i<height; i++){
	  for(int j=0; j<width; j++){	    

	    //first we considere the finite temperature case.
	    if ((temperature>0) && (i==j)){   

	      double rescaled_temperature = temperature*pow (lambda, 0.5*((double)(iteration_no)));
	      
	      /* the Hamiltonian is already diagonal... */
	      double energy = new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j]; 

	      double matrix_element = (1.0/partition_function)*exp(-(energy-energy_min)/rescaled_temperature);
	 
	     
	      new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i,j, matrix_element); 
    
	      //    cout<< rescaled_temperature<<"\t"<<energy_min<<"\t"<<energy<<"\t"<<matrix_element<<endl;
	      // getchar();

	    }//if temperature>0


	    else{

	      //fill with zero's all the matrix elements at T=0 temperature

	      new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i,j, 0.0); 	    

	    }//else	   

	  }//for j
	}//for i
      }//if grand_selection_rule      
    }//for sector_column;
  }//for sector_row
    
  if(temperature==0){
    
    
    // determination of the degeneracy of the ground_state 

    int no_ground_states = 0;
    
    for(int e_index = 0; e_index < new_block_state_no; e_index++){
      if(fabs(energy_min - new_block_energies_array[e_index]) <degeneracy_threshhold){
	no_ground_states++;      
      }//if
    }//for e_index
     
    /* first count the total ground state degeneracy */ 

    double total_degeneracy = 0.0;
    for(int ground_state_index = 0; ground_state_index < no_ground_states; ground_state_index++){
	/* get index of this ground state */ 
	int energy_min_index = new_block_energies_index[ground_state_index] - 1;      
	
	/* get sector of this ground state */ 
	int ground_state_sector = new_block_states_array[energy_min_index].get_sector();
	
	total_degeneracy += (double) new_block_sectors_properties_array[ground_state_sector].degeneracy;
     }//for ground_state_index;

    /* now I fill up the T=0 density matrix */ 

    for(int ground_state_index = 0; ground_state_index < no_ground_states; ground_state_index++){
      
    // We determine to which sector (ground_sect_ind) and in that sector
    //   to which state (ground_state_ind) does the ground state energy belong.
      
	/* get index of this ground state */ 
      int energy_min_index = new_block_energies_index[ground_state_index] - 1;      

      /* get sector of this ground state */ 
      int ground_state_sector = new_block_states_array[energy_min_index].get_sector();

      /* get index inside the sector of this ground state */ 
      int ground_state_sector_index = energy_min_index - new_block_sectors_properties_array[ground_state_sector].first_state;            
      
      double matrix_element = 1.0/ total_degeneracy;

      new_block_density_matrix.get_sector_matrix()[ground_state_sector][ground_state_sector].
	set_matrix_element(ground_state_sector_index, ground_state_sector_index, matrix_element); 
 
    }//for ground_state_index; 

  }// if temperature=0

  
  return; 
}//construct_density_matrix


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_m_star(int &m_star,
		 char *m_star_file_name){

/*******************************************************************************
input variables:  one
     m_star_file_name - the file name which contains the value of m_star

output variables: one
     m_star - the iteration index when truncation of the block states first took 
              place.

return: void

What it does:
     This routine reads m_star from given by file m_star_file_name.
How to call:
*******************************************************************************/


  m_star_swap_file_name(swap_folder_name,
			m_star_file_name); 

  ifstream m_star_file(m_star_file_name, ios::in); 
  
  if (!m_star_file.good()){
      error_message("reading m_star"); 
  }//if

  m_star_file.getline(buffer, buffer_size, '\n');
  word_offset = 0; 

  if(get_word(buffer, word, word_offset)){
    if(strcmp(word, "m_star")==0){
      if(!get_word(buffer, word, word_offset)){
	error_message("m_star in routine read_m_star"); 
      }//if
      m_star= atoi(word); 	      	
    }//if 
  }//if 
  
  

  m_star_file.close();
  m_star_file.clear();
  return; 
}


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_states(int m_star,
		 long iteration_index, 

		 int symmetry_no,

		 int & new_block_state_no,
		 state * & new_block_states_array,
		 double * & new_block_energies_array,
		 double * & sorted_new_block_energies_array,
		 unsigned long * & new_block_energies_index,

		 int &block_state_no, 
		 state * & block_states_array

		 ){

/*************************************************************************
input variables:  three
    m_star          - the iteration at which the writing process begins. 
    iteration_index - the index of the current iteration. This variable is 
                      written in the file name. 
    symmtry_no      - is the number of symmetries in the problem. 

output variables: one
    new_block_states_array - array with the states (readed from the file, 
                       all the variables that describe a state are readed 
                       from the file)

    new_block_states_energies_array - array with energies of the states 
                                (readed from the file)
    sorted_new_block_energies_array - array with the energies of the states sorted
                                in ascending order.  
                                computed once the energies are readed
    new_block_energies_index        - array that indexes the order of the states 
                                energies. ( computed ones the energies are 
                                readed. )
    block_state_no          - the number of current states. (readed from 
                                the file)
    block_states_array      - array with the block_states (all the variables 
                                that describes the state class are readed 
                                from the file)

return: bool

What it does:
    This routine read all the states at a given iteration from the corresponding
    file. this file was created previosuly in the course of the NRG 
    iteration. 

How to call:
*************************************************************************/


  if (m_star ==-1){
    cout<<"Error in routine read_states(), m_star=-1\nEXITING...\n"; 
  }//if m_star


  if(new_block_states_array){
    delete [] new_block_states_array; 
    new_block_states_array =NULL;
  }
  if (new_block_energies_array){
    delete [] new_block_energies_array; 
    new_block_energies_array =NULL;
  }
  if (sorted_new_block_energies_array){
    delete [] sorted_new_block_energies_array;
    sorted_new_block_energies_array =NULL;
  }

  if (new_block_energies_index){
    delete [] new_block_energies_index; 
    new_block_energies_index = NULL;
  }
  
  if (block_states_array){
    delete [] block_states_array; 
    block_states_array = NULL; 
  }


  if((iteration_index >= m_star) && (m_star >= 0)){
    
    states_swap_file_name(states_folder_name, 
			  iteration_index,
			  states_file_name
			  );  
 
    states_file_in.open(states_file_name, ios::in);
   
    if (!states_file_in.good()){
      error_message("reading states_file"); 
    }
   
    states_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "new_block_state_no")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("new_block_state_no in routine read_states"); 
	    }//if
	    new_block_state_no= atoi(word); 	      	
      }//if 
    }//if 

    //this is the commment line
    states_file_in.getline(buffer, buffer_size, '\n');


    //now we create the new_block_states


    new_block_states_array = new state [new_block_state_no]; 
    new_block_energies_array = new double [new_block_state_no];  
    new_block_energies_index = new unsigned long [new_block_state_no]; 
    sorted_new_block_energies_array = new double [new_block_state_no]; 


    
    for(int state_index = 0; state_index < new_block_state_no; state_index++){

      states_file_in.getline(buffer, buffer_size, '\n'); 
      

      word_offset = 0; 
      if(!get_word(buffer, word, word_offset)){
	error_message("reading sector"); 
      }//if
      new_block_states_array[state_index].set_sector(atoi(word)); 
           

      if(!get_word(buffer, word, word_offset)){
	error_message("reading sector_index"); 
      }//if
      
      new_block_states_array[state_index].set_sector_index(atoi(word)); 


      new_block_states_array[state_index].activate(symmetry_no); 

      for(int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
	if(!get_word(buffer, word, word_offset)){
	  error_message("reading representation_index"); 
	}//if
	new_block_states_array[state_index].set_representation_index(symmetry_index, atoi(word)); 
      }//for symmetry_index

      

      if(!get_word(buffer, word, word_offset)){
	error_message("reading parent_state_index"); 
      }//if
      new_block_states_array[state_index].set_parent_state_index(atoi(word)); 
      
      if(!get_word(buffer, word, word_offset)){
	error_message("reading local_state_index"); 
      }//if
      new_block_states_array[state_index].set_local_state_index(atoi(word)); 

      if(!get_word(buffer, word, word_offset)){
	error_message("reading sgn"); 
      }//if
      new_block_states_array[state_index].set_sgn(atoi(word)); 

      if(!get_word(buffer, word, word_offset)){
	error_message("reading elimination_tag"); 
      }//if
      new_block_states_array[state_index].set_elimination_tag(word[0]); 
      assert((new_block_states_array[state_index].get_elimination_tag()=='K')||(new_block_states_array[state_index].get_elimination_tag()=='D')); 
      
      if(!get_word(buffer, word, word_offset)){
	error_message("reading new_block_energies_array"); 
      }//if
      new_block_energies_array[state_index] = atof(word);       	     	     
    }//for state_index


    //we sort the new_block_energies_array here.
    
    generate_energy_index_table(new_block_state_no, new_block_energies_array, new_block_energies_index); 
    sort_energies(new_block_state_no, new_block_energies_array, sorted_new_block_energies_array); 


    states_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "block_state_no")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("block_state_no in routine read_states"); 
	    }//if
	    block_state_no= atoi(word); 	      	
      }//if 
    }//if 

  //this is the commment line
    states_file_in.getline(buffer, buffer_size, '\n');


    block_states_array = new state [block_state_no]; 


    for(int state_index = 0; state_index < block_state_no; state_index++){

      states_file_in.getline(buffer, buffer_size, '\n'); 

      
      word_offset = 0; 
      if(!get_word(buffer, word, word_offset)){
	error_message("reading sector"); 
      }//if
      block_states_array[state_index].set_sector(atoi(word)); 
           

      if(!get_word(buffer, word, word_offset)){
	error_message("reading sector_index"); 
      }//if
      
      block_states_array[state_index].set_sector_index(atoi(word)); 


      block_states_array[state_index].activate(symmetry_no); 

      for(int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
	if(!get_word(buffer, word, word_offset)){
	  error_message("reading representation_index"); 
	}//if
	block_states_array[state_index].set_representation_index(symmetry_index, atoi(word)); 
      }//for symmetry_index

      

      if(!get_word(buffer, word, word_offset)){
	error_message("reading parent_state_index"); 
      }//if
      block_states_array[state_index].set_parent_state_index(atoi(word)); 
      
      if(!get_word(buffer, word, word_offset)){
	error_message("reading local_state_index"); 
      }//if
      block_states_array[state_index].set_local_state_index(atoi(word)); 

      if(!get_word(buffer, word, word_offset)){
	error_message("reading sgn"); 
      }//if
      block_states_array[state_index].set_sgn(atoi(word)); 

      if(!get_word(buffer, word, word_offset)){
	error_message("reading elimination_tag"); 
      }//if
      block_states_array[state_index].set_elimination_tag(word[0]); 
      assert((block_states_array[state_index].get_elimination_tag()=='K')||(block_states_array[state_index].get_elimination_tag()=='D')); 
      
    }//for state_index
        
    states_file_in.close();
    states_file_in.clear();
   
  }//if  
  return; 
}//read_states


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_block_states(int m_star,
		       long iteration_index, 
		       
		       int symmetry_no,
		       
		       int &block_state_no, 
		       state * & block_states_array,
		       
		       int * & block_original_sector_array, 
		       int * & block_original_sector_index_array
			 
		       ){

/*************************************************************************
input variables:  three
     m_star          - the iteration at which the writing process begins. 
     iteration_index - the index of the current iteration. This variable is 
                       written in the file name. 
     symmtry_no      - is the number of symmetries in the problem. 

output variables: two
     block_state_no     - the number of current states. (readed from the file)
     block_states_array - array with the block_states (all the variables 
                            that describes the state class are readed from the 
                            file)

return: void

What it does:
     This routine read only the current  states at a given iteration from the 
     corresponding
     file. this file was created previosuly in the course of the NRG 
     iteration. The file contains the new_block_states_array as well as the 
     block_states_array. 

How to call:
*************************************************************************/
  
  int new_block_state_no; 

  if (m_star ==-1){
    cout<<"Error in routine read_states(), m_star=-1\nEXITING...\n"; 
  }//if m_star

  
  if (block_states_array){
    delete [] block_states_array; 
    block_states_array = NULL; 
  }

  if (block_original_sector_array){
    delete [] block_original_sector_array; 
  }

  block_original_sector_array = NULL;

  if (block_original_sector_index_array){
    delete [] block_original_sector_index_array; 
  }

  block_original_sector_index_array = NULL;


  if((iteration_index >= m_star) && (m_star >= 0)){
    
    states_swap_file_name(states_folder_name, 
			  iteration_index,
			  states_file_name
			  );  
 
    states_file_in.open(states_file_name, ios::in);
   
    if (!states_file_in.good()){
      error_message("reading states_file"); 
    }
   
    states_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "new_block_state_no")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("new_block_state_no in routine read_block_states_array"); 
	    }//if
	    new_block_state_no= atoi(word); 	      	
      }//if 
    }//if 

    //this is the commment line
    states_file_in.getline(buffer, buffer_size, '\n');

    //now we create the new_block_states_array
    
    for(int state_index = 0; state_index < new_block_state_no; state_index++){

      states_file_in.getline(buffer, buffer_size, '\n'); 
    }
      

    states_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "block_state_no")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("block_state_no in routine read_states"); 
	    }//if
	    block_state_no= atoi(word); 	      	
      }//if 
    }//if 

  //this is the commment line
    states_file_in.getline(buffer, buffer_size, '\n');


    block_states_array = new state [block_state_no]; 

    block_original_sector_array = new int [block_state_no];
    block_original_sector_index_array = new int [block_state_no];


    for(int state_index = 0; state_index < block_state_no; state_index++){

      states_file_in.getline(buffer, buffer_size, '\n'); 

      
      word_offset = 0; 
      if(!get_word(buffer, word, word_offset)){
	error_message("reading sector"); 
      }//if
      block_states_array[state_index].set_sector(atoi(word)); 
           

      if(!get_word(buffer, word, word_offset)){
	error_message("reading sector_index"); 
      }//if
      
      block_states_array[state_index].set_sector_index(atoi(word)); 


      block_states_array[state_index].activate(symmetry_no); 

      for(int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
	if(!get_word(buffer, word, word_offset)){
	  error_message("reading representation_index"); 
	}//if
	block_states_array[state_index].set_representation_index(symmetry_index, atoi(word)); 
      }//for symmetry_index

      

      if(!get_word(buffer, word, word_offset)){
	error_message("reading parent_state_index"); 
      }//if
      block_states_array[state_index].set_parent_state_index(atoi(word)); 
      
      if(!get_word(buffer, word, word_offset)){
	error_message("reading local_state_index"); 
      }//if
      block_states_array[state_index].set_local_state_index(atoi(word)); 

      if(!get_word(buffer, word, word_offset)){
	error_message("reading sgn"); 
      }//if
      block_states_array[state_index].set_sgn(atoi(word)); 

      if(!get_word(buffer, word, word_offset)){
	error_message("reading elimination_tag"); 
      }//if
      block_states_array[state_index].set_elimination_tag(word[0]); 
      assert((block_states_array[state_index].get_elimination_tag()=='K')||(block_states_array[state_index].get_elimination_tag()=='D')); 
      
      if(!get_word(buffer, word, word_offset)){
	error_message("reading block_original_sector_array"); 
      }//if
      block_original_sector_array[state_index] =atoi(word); 
      
      if(!get_word(buffer, word, word_offset)){
	error_message("reading block_original_sector_index_array"); 
      }//if
      block_original_sector_index_array[state_index] =atoi(word); 
      
      

    }//for state_index
        
    states_file_in.close();
    states_file_in.clear();
   
  }//if  
  return; 
}//read_block_states


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_new_block_states(int m_star,
		     long iteration_index, 
		     
		     int symmetry_no,
		     
		     int & new_block_state_no,
		     state * & new_block_states_array,
		     double * & new_block_energies_array,
		     double * & sorted_new_block_energies_array,
		     unsigned long * & new_block_energies_index
		     ){

/*************************************************************************
input variables:  three
     m_star          - the iteration at which the writing process begins. 
     iteration_index - the index of the current iteration. This variable is 
                       written in the file name. 
     symmtry_no      - is the number of symmetries in the problem. 

output variables: one
     new_block_states_array - array with the states (readed from the file, 
                        all the variables that describe a state are readed 
                        from the file)

     new_block_states_energies_array - array with energies of the states (readed from
                                 the file)
     sorted_new_block_energies_array - array with the energies of the states sorted 
                                 in ascending order.  
                                 computed once the energies are readed
     new_block_energies_index        - array that indexes the order of the states 
                                 energies. ( computed ones the energies are 
                                 readed. )

return: void

What it does:
     This routine reads only the new  states at a given iteration from the 
     corresponding
     file. this file was created previosuly in the course of the NRG 
     iteration. The file contains also the block_states_array 
     but these are not readed with the present routine.

How to call:
*************************************************************************/

  if (m_star ==-1){
    cout<<"Error in routine read_new_block_states(), m_star=-1\nEXITING...\n"; 
  }//if m_star



  if(new_block_states_array){
    delete [] new_block_states_array; 
    new_block_states_array =NULL;
  }
  if (new_block_energies_array){
    delete [] new_block_energies_array; 
    new_block_energies_array =NULL;
  }
  if (sorted_new_block_energies_array){
    delete [] sorted_new_block_energies_array;
    sorted_new_block_energies_array =NULL;
  }

  if (new_block_energies_index){
    delete [] new_block_energies_index; 
    new_block_energies_index = NULL;
  }
  

  if((iteration_index >= m_star) && (m_star >= 0)){
    
    states_swap_file_name(states_folder_name, 
			  iteration_index,
			  states_file_name
			  );  
 
    states_file_in.open(states_file_name, ios::in);
   
    if (!states_file_in.good()){
      error_message("reading states_file"); 
    }
   
    states_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "new_block_state_no")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("new_block_state_no in routine read_states"); 
	    }//if
	    new_block_state_no= atoi(word); 	      	
      }//if 
    }//if 

    //this is the commment line
    states_file_in.getline(buffer, buffer_size, '\n');


    //now we create the new_block_states_array


    new_block_states_array = new state [new_block_state_no]; 
    new_block_energies_array = new double [new_block_state_no];  
    new_block_energies_index = new unsigned long [new_block_state_no]; 
    sorted_new_block_energies_array = new double [new_block_state_no]; 


    
    for(int state_index = 0; state_index < new_block_state_no; state_index++){

      states_file_in.getline(buffer, buffer_size, '\n'); 

      
      word_offset = 0; 
      if(!get_word(buffer, word, word_offset)){
	error_message("reading sector"); 
      }//if
      new_block_states_array[state_index].set_sector(atoi(word)); 
           

      if(!get_word(buffer, word, word_offset)){
	error_message("reading sector_index"); 
      }//if
      
      new_block_states_array[state_index].set_sector_index(atoi(word)); 


      new_block_states_array[state_index].activate(symmetry_no); 

      for(int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
	if(!get_word(buffer, word, word_offset)){
	  error_message("reading representation_index"); 
	}//if
	new_block_states_array[state_index].set_representation_index(symmetry_index, atoi(word)); 
      }//for symmetry_index

      

      if(!get_word(buffer, word, word_offset)){
	error_message("reading parent_state_index"); 
      }//if
      new_block_states_array[state_index].set_parent_state_index(atoi(word)); 
      
      if(!get_word(buffer, word, word_offset)){
	error_message("reading local_state_index"); 
      }//if
      new_block_states_array[state_index].set_local_state_index(atoi(word)); 

      if(!get_word(buffer, word, word_offset)){
	error_message("reading sgn"); 
      }//if
      new_block_states_array[state_index].set_sgn(atoi(word)); 

      if(!get_word(buffer, word, word_offset)){
	error_message("reading elimination_tag"); 
      }//if
      new_block_states_array[state_index].set_elimination_tag(word[0]); 
      assert((new_block_states_array[state_index].get_elimination_tag()=='K')||(new_block_states_array[state_index].get_elimination_tag()=='D')); 
      
      if(!get_word(buffer, word, word_offset)){
	error_message("reading new_block_energies_array"); 
      }//if
      new_block_energies_array[state_index] = atof(word);       	     	     
    }//for state_index


    //we sort the energies here.
    
    generate_energy_index_table(new_block_state_no, new_block_energies_array, new_block_energies_index); 
    sort_energies(new_block_state_no, new_block_energies_array, sorted_new_block_energies_array); 


    states_file_in.close();
    states_file_in.clear();
   
  }//if  
  return; 
}//read_new_block_states


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_transformation_sectors(int m_star,
				 long iteration_index, 
				 int new_block_state_no,
				 int new_block_sector_no, 
				 sector_properties *new_block_sectors_properties_array,
				 sector * & transformation_sectors
				 ){

/*************************************************************************
input variables:  two
     m_star          - the iteration at which the writing process begins. 
     iteration_index - the index of the current iteration. This variable is 
                       written in the file name. 
     new_block_state_no    - the number of new states. 
     new_block_sector_no   - the number of sectors
     new_block_sectors_properties_array - array with the  properties of the given 
                                    sectors. 

output variables: one
     transformation_sectors - array with the transformation sectors that were
     obtained previously during the forward NRG iteration.

return: none

What it does:
     This routine reads  the transformation sectors  at a given iteration from 
     the corresponding
     file. this file was created previosuly in the course of the NRG 
     iteration. All the data in the file is readed at once. First the states has to be 
     read in, 
     and the sectors have to be set up before reading the transformation_sectors. 

How to call:
*************************************************************************/
 
  const int buffer_size = 262143;
  char buffer[buffer_size + 1];
  char word[buffer_size + 1];
  int word_offset;


  if (m_star ==-1){
    cout<<"Error in routine read_states(), m_star=-1\nEXITING...\n"; 
  }//if m_star


  if (transformation_sectors){
    delete [] transformation_sectors; 
    transformation_sectors = NULL; 
  }

  
  transformation_sectors = new sector [new_block_sector_no]; 

  for(int sector_index = 0; sector_index < new_block_sector_no; sector_index++){
        
    int sector_size = new_block_sectors_properties_array[sector_index].dimension;
    transformation_sectors[sector_index].set_size(sector_size, sector_size);    
    transformation_sectors[sector_index].allocate_matrix();
  }//sector_index; 




  if((iteration_index >= m_star) && (m_star >= 0)){

    transformation_swap_file_name( transformations_folder_name, 
				   iteration_index,
				   transformation_file_name
				   );   

    transformation_file_in.open(transformation_file_name, ios::in);
   
    if (!transformation_file_in.good()){
      error_message("reading transformation_file"); 
    }
  
    transformation_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "new_block_sector_no")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("new_block_sector_no in routine read_states"); 
	    }//if
	    assert (new_block_sector_no == atoi(word));
      }//if 
    }//if 
    
    for (int sector_index = 0; sector_index<new_block_sector_no; sector_index++){

      transformation_file_in.getline(buffer, buffer_size, '\n');
      word_offset = 0; 
      
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "sector")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("sector in routine read_states"); 
	    }//if
	    assert (sector_index == atoi(word));
      }//if 
    }//if 

      transformation_file_in.getline(buffer, buffer_size, '\n');
      word_offset = 0; 
      
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "sector_dimension")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("sector_dimension in routine read_states"); 
	    }//if
	    assert (new_block_sectors_properties_array[sector_index].dimension == atoi(word));
      }//if 
    }//if 

      int sector_dimension = new_block_sectors_properties_array[sector_index].dimension; 

      for (int i=0; i<sector_dimension; i++){

	transformation_file_in.getline(buffer, buffer_size, '\n');
	word_offset = 0; 

	for(int j=0; j<sector_dimension; j++){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading transformation_matrix"); 
	  }//if
	  transformation_sectors[sector_index].set_matrix_element(i, j, atof(word));
	}//j
      }//i     
    }//sector_index


    transformation_file_in.close();
    transformation_file_in.clear();
  }//if iteration_index>=m_star

  return; 
}//read_transformation_sectors




/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_transformation_sectors_binary(int m_star,
					long iteration_index, 
					int new_block_state_no,
					int new_block_sector_no, 
					sector_properties *new_block_sectors_properties_array,
					sector * & transformation_sectors
					){

/*************************************************************************
input variables:  two
     m_star          - the iteration at which the writing process begins. 
     iteration_index - the index of the current iteration. This variable is 
                       written in the file name. 
     new_block_state_no    - the number of new states. 
     new_block_sector_no   - the number of sectors
     new_block_sectors_properties_array - array with the  properties of the given 
                                    sectors. 

output variables: one
     transformation_sectors - array with the transformation sectors that were
     obtained previously during the forward NRG iteration.

return: none

What it does:
     This routine reads  the transformation sectors  at a given iteration from 
     the corresponding
     file. this file was created previosuly in the course of the NRG 
     iteration. All the data in the file is readed at once. First the states has to be 
     read in, 
     and the sectors have to be set up before reading the transformation_sectors. 

How to call:
*************************************************************************/

  if (m_star ==-1){
    cout<<"Error in routine read_states(), m_star=-1\nEXITING...\n"; 
  }//if m_star

  
  transformation_sectors = new sector [new_block_sector_no]; 

  for(int sector_index = 0; sector_index < new_block_sector_no; sector_index++){
        
    int sector_size = new_block_sectors_properties_array[sector_index].dimension;
    transformation_sectors[sector_index].set_size(sector_size, sector_size);    
    transformation_sectors[sector_index].allocate_matrix();
  }//sector_index; 




  if((iteration_index >= m_star) && (m_star >= 0)){

    transformation_binary_swap_file_name( transformations_folder_name, 
					  iteration_index,
					  transformation_file_name
					  );   

    transformation_file_in.open(transformation_file_name, ios::in| ios::binary);
   
    if (!transformation_file_in.good()){
      error_message("reading transformation_file"); 
    }


    int sector_no; 

    transformation_file_in.read((char *) &sector_no, sizeof(int));
    assert(new_block_sector_no==sector_no); 

    for(int sector_label = 0; sector_label < new_block_sector_no; sector_label++){

      transformation_file_in.read(( char *)&sector_label, sizeof(int));

      int sector_dimension; 
      transformation_file_in.read((char *) &sector_dimension, sizeof(int)); 
      assert(sector_dimension==new_block_sectors_properties_array[sector_label].dimension); 

      double matrix_element; 
      for (int i=0; i<sector_dimension; i++){
	for(int j=0; j<sector_dimension; j++){
	  transformation_file_in.read(( char *) &matrix_element, sizeof(double)); 
	  transformation_sectors[sector_label].set_matrix_element(i, j, matrix_element);
	}//j
      }//i

    }//for sector_label
    transformation_file_in.close();
    transformation_file_in.clear();
  }//if iteration_index>=m_star

  return; 
}//read_transformation_sectors_binary



/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_spectral_operator(int m_star, 
			    int iteration_index, 
			    int new_block_state_no, 
			    int new_block_sector_no, 
			    sector_properties * new_block_sectors_properties_array, 
			    quantum_operator & new_block_spectral_operator){

/*************************************************************************
input variables:  two
    m_star          - the iteration at which the writing process begins. 
    iteration_index - the index of the current iteration. This variable is 
                      written in the file name. 
    new_block_state_no    - the number of new states. 
    new_block_sector_no   - the number of sectors
    new_block_sectors_properties_array - array with the  properties of the given 
                                   sectors. 

output variables: one
    new_block_spectral_operator - the corresponding operator readed from the file. 
                            the corresponding sector matrices of this operator
                            were obtained in the course of forward NRG 
                            iteration. 

return: bool

What it does:
     This routine reads  one spectral operator  at a given iteration from the 
     corresponding
     file. this file was created previosuly in the course of the NRG 
     iteration. All the data in the file is readed at once. First the states 
     has to be read in, 
     and the sectors have to be set up before reading the transformation_sectors. 
     The name of the file contains the name of the spectral_operator and the 
     iteration index. 

How to call:
*************************************************************************/
  
  const int buffer_size = 262143;
  char buffer[buffer_size + 1];
  char word[buffer_size + 1];
  int word_offset;


  if (m_star ==-1){
    cout<<"Error in routine read_spectral_operator(), m_star=-1\nEXITING...\n"; 
  }//if m_star

  
  if((iteration_index >= m_star) && (m_star >= 0)){

    new_block_spectral_operator.clear_sectors_matrix();
    new_block_spectral_operator.allocate_sectors_matrix(new_block_sector_no); 


    for(int sector_row=0; sector_row<new_block_sector_no; sector_row++){
      for(int sector_column=0; sector_column<new_block_sector_no; sector_column++){
	
	
	new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
										      new_block_sectors_properties_array[sector_column].dimension);


	if(grand_selection_rule(symmetry_no, 
				symmetry_types, 
				new_block_sectors_properties_array[sector_column].representation_index,
				new_block_spectral_operator.get_representation_index(), 
				new_block_sectors_properties_array[sector_row].representation_index)){
	  
	  

	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].allocate_matrix();
	  
	}//if grand_selection_rule
	  
      }//for sector_column
    }//for sector_row



      

    spectral_operator_swap_file_name( spectral_operators_folder_name, 
				      iteration_index,
				      spectral_operator_file_name,
				      new_block_spectral_operator
				      );  
    
    spectral_operator_file_in.open(spectral_operator_file_name, ios::in); 
    
    if (!spectral_operator_file_in.good()){
      error_message("reading spectral_operator_file_in"); 
    }//if

    
    spectral_operator_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    int active_sector_no;
 
    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "active_sector_no")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("reading active_sector_no"); 
	    }//if
	    active_sector_no= atoi(word); 	      	
      }//if 
    }//if 

    
    for(int active_sector_index=0; active_sector_index<active_sector_no; active_sector_index++){

      int sector_row; 
      int sector_column; 
      int row_dimension;
      int column_dimension; 
      double matrix_element; 

      spectral_operator_file_in.getline(buffer, buffer_size, '\n'); 
      word_offset =0; 
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "sector_row")==0){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading sector_row"); 
	  }//if
	  sector_row= atoi(word); 	      	
	}//if 
      }//if 
      
      spectral_operator_file_in.getline(buffer, buffer_size, '\n'); 
      word_offset =0; 
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "sector_column")==0){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading sector_column"); 
	  }//if
	  sector_column= atoi(word); 	      	
	}//if        
      }//if 
      
      spectral_operator_file_in.getline(buffer, buffer_size, '\n'); 
      word_offset =0; 
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "row_dimension")==0){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading row_dimension"); 
	  }//if
	  row_dimension= atoi(word); 	      	
	}//if        
      }//if 
      
      spectral_operator_file_in.getline(buffer, buffer_size, '\n'); 
      word_offset =0; 
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "column_dimension")==0){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading column_dimension"); 
	  }//if
	  column_dimension= atoi(word); 	      	
	}//if        
      }//if 

      
      assert (new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].get_height() == row_dimension); 
      assert (new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].get_width() == column_dimension); 

      for (int i=0; i<row_dimension; i++){
	spectral_operator_file_in.getline(buffer, buffer_size, '\n'); 
	word_offset =0; 

	for(int j=0; j<column_dimension; j++){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading new_block_spectral_operator sector_matrix."); 
	  }//if

	  matrix_element = atof(word); 
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i, j, matrix_element);             	       
	}//j
      }//for i
    }//for active_sector_index
    
  spectral_operator_file_in.close();
  spectral_operator_file_in.clear();
  }//if iteration_index>=m_star

  return;

}//read_spectral_operator;




/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_spectral_operator_binary(int m_star, 
				   int iteration_index, 
				   int new_block_state_no, 
				   int new_block_sector_no, 
				   sector_properties * new_block_sectors_properties_array, 
				   quantum_operator & new_block_spectral_operator){
  
/*************************************************************************
input variables:  two
    m_star          - the iteration at which the writing process begins. 
    iteration_index - the index of the current iteration. This variable is 
                      written in the file name. 
    new_block_state_no    - the number of new states. 
    new_block_sector_no   - the number of sectors
    new_block_sectors_properties_array - array with the  properties of the given 
                                   sectors. 

output variables: one
    new_block_spectral_operator - the corresponding operator readed from the file. 
                            the corresponding sector matrices of this operator
                            were obtained in the course of forward NRG 
                            iteration. 

return: bool

What it does:
     This routine reads  one spectral operator  at a given iteration from the 
     corresponding
     file. this file was created previosuly in the course of the NRG 
     iteration. All the data in the file is readed at once. First the states 
     has to be read in, 
     and the sectors have to be set up before reading the transformation_sectors. 
     The name of the file contains the name of the spectral_operator and the 
     iteration index. 

How to call:
*************************************************************************/
  
  if (m_star ==-1){
    cout<<"Error in routine read_spectral_operator(), m_star=-1\nEXITING...\n"; 
  }//if m_star

  
  if((iteration_index >= m_star) && (m_star >= 0)){

    new_block_spectral_operator.clear_sectors_matrix();
    new_block_spectral_operator.allocate_sectors_matrix(new_block_sector_no); 


    for(int sector_row=0; sector_row<new_block_sector_no; sector_row++){
      for(int sector_column=0; sector_column<new_block_sector_no; sector_column++){
	
	
	new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
										      new_block_sectors_properties_array[sector_column].dimension);


	if(grand_selection_rule(symmetry_no, 
				symmetry_types, 
				new_block_sectors_properties_array[sector_column].representation_index,
				new_block_spectral_operator.get_representation_index(), 
				new_block_sectors_properties_array[sector_row].representation_index)){
	  
	  

	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].allocate_matrix();
	  
	}//if grand_selection_rule
	  
      }//for sector_column
    }//for sector_row



      

    spectral_operator_binary_swap_file_name( spectral_operators_folder_name, 
					     iteration_index,
					     spectral_operator_file_name,
					     new_block_spectral_operator
					     );  
    
    spectral_operator_file_in.open(spectral_operator_file_name, ios::in|ios::binary); 
    
    if (!spectral_operator_file_in.good()){
      error_message("reading spectral_operator_file_in"); 
    }//if

    
    int active_sector_no;

    spectral_operator_file_in.read(( char *) &active_sector_no, sizeof(int));

    for(int active_sector_index=0; active_sector_index<active_sector_no; active_sector_index++){

      int sector_row; 
      int sector_column; 
      int row_size;
      int column_size; 

      spectral_operator_file_in.read(( char *) &sector_row, sizeof(int));
      spectral_operator_file_in.read(( char *) &sector_column, sizeof(int));;
	
      spectral_operator_file_in.read(( char *) &row_size, sizeof(int)); 
      spectral_operator_file_in.read(( char *) &column_size, sizeof(int)); 
	
      assert (new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].get_height() == row_size); 
      assert (new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].get_width() == column_size); 

	
      double matrix_element; 
      for (int i=0; i<row_size; i++){
	for(int j=0;j<column_size; j++){
	  spectral_operator_file_in.read(( char *) &matrix_element, sizeof(double)); 
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i, j, matrix_element);
	}//j
      }//i
	
    }//for active_sector_index

    spectral_operator_file_in.close();
    spectral_operator_file_in.clear();

  }//if iteration_index>=m_star
  
  return;
  
}//read_spectral_operator_binary;






/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_static_operator(int m_star, 
			    int iteration_index, 
			    int new_block_state_no, 
			    int new_block_sector_no, 
			    sector_properties * new_block_sectors_properties_array, 
			    quantum_operator & new_block_static_operator){

/*************************************************************************
input variables:  two
    m_star          - the iteration at which the writing process begins. 
    iteration_index - the index of the current iteration. This variable is 
                      written in the file name. 
    new_block_state_no    - the number of new states. 
    new_block_sector_no   - the number of sectors
    new_block_sectors_properties_array - array with the  properties of the given 
                                   sectors. 

output variables: one
    new_block_static_operator - the corresponding operator readed from the file. 
                            the corresponding sector matrices of this operator
                            were obtained in the course of forward NRG 
                            iteration. 

return: bool

What it does:
     This routine reads  one static operator  at a given iteration from the 
     corresponding
     file. this file was created previosuly in the course of the NRG 
     iteration. All the data in the file is readed at once. First the states 
     has to be read in, 
     and the sectors have to be set up before reading the transformation_sectors. 
     The name of the file contains the name of the static_operator and the 
     iteration index. 

How to call:
*************************************************************************/
  
  const int buffer_size = 262143;
  char buffer[buffer_size + 1];
  char word[buffer_size + 1];
  int word_offset;


  if (m_star ==-1){
    cout<<"Error in routine read_static_operator(), m_star=-1\nEXITING...\n"; 
  }//if m_star

  
  if((iteration_index >= m_star) && (m_star >= 0)){

    new_block_static_operator.clear_sectors_matrix();
    new_block_static_operator.allocate_sectors_matrix(new_block_sector_no); 


    for(int sector_row=0; sector_row<new_block_sector_no; sector_row++){
      for(int sector_column=0; sector_column<new_block_sector_no; sector_column++){
	
	
	new_block_static_operator.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
										      new_block_sectors_properties_array[sector_column].dimension);


	if(grand_selection_rule(symmetry_no, 
				symmetry_types, 
				new_block_sectors_properties_array[sector_column].representation_index,
				new_block_static_operator.get_representation_index(), 
				new_block_sectors_properties_array[sector_row].representation_index)){
	  
	  

	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].allocate_matrix();
	  
	}//if grand_selection_rule
	  
      }//for sector_column
    }//for sector_row



      

    static_operator_swap_file_name( static_operators_folder_name, 
				      iteration_index,
				      static_operator_file_name,
				      new_block_static_operator
				      );  
    
    static_operator_file_in.open(static_operator_file_name, ios::in); 
    
    if (!static_operator_file_in.good()){
      error_message("reading static_operator_file_in"); 
    }//if

    
    static_operator_file_in.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    int active_sector_no;
 
    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "active_sector_no")==0){
	    if(!get_word(buffer, word, word_offset)){
	      error_message("reading active_sector_no"); 
	    }//if
	    active_sector_no= atoi(word); 	      	
      }//if 
    }//if 

    
    for(int active_sector_index=0; active_sector_index<active_sector_no; active_sector_index++){

      int sector_row; 
      int sector_column; 
      int row_dimension;
      int column_dimension; 
      double matrix_element; 

      static_operator_file_in.getline(buffer, buffer_size, '\n'); 
      word_offset =0; 
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "sector_row")==0){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading sector_row"); 
	  }//if
	  sector_row= atoi(word); 	      	
	}//if 
      }//if 
      
      static_operator_file_in.getline(buffer, buffer_size, '\n'); 
      word_offset =0; 
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "sector_column")==0){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading sector_column"); 
	  }//if
	  sector_column= atoi(word); 	      	
	}//if        
      }//if 
      
      static_operator_file_in.getline(buffer, buffer_size, '\n'); 
      word_offset =0; 
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "row_dimension")==0){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading row_dimension"); 
	  }//if
	  row_dimension= atoi(word); 	      	
	}//if        
      }//if 
      
      static_operator_file_in.getline(buffer, buffer_size, '\n'); 
      word_offset =0; 
      if(get_word(buffer, word, word_offset)){
	if(strcmp(word, "column_dimension")==0){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading column_dimension"); 
	  }//if
	  column_dimension= atoi(word); 	      	
	}//if        
      }//if 

      
      assert (new_block_static_operator.get_sector_matrix()[sector_row][sector_column].get_height() == row_dimension); 
      assert (new_block_static_operator.get_sector_matrix()[sector_row][sector_column].get_width() == column_dimension); 

      for (int i=0; i<row_dimension; i++){
	static_operator_file_in.getline(buffer, buffer_size, '\n'); 
	word_offset =0; 

	for(int j=0; j<column_dimension; j++){
	  if(!get_word(buffer, word, word_offset)){
	    error_message("reading new_block_static_operator sector_matrix."); 
	  }//if

	  matrix_element = atof(word); 
	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i, j, matrix_element);             	       
	}//j
      }//for i
    }//for active_sector_index
    
  static_operator_file_in.close();
  static_operator_file_in.clear();
  }//if iteration_index>=m_star

  return;

}//read_static_operator;




/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void read_static_operator_binary(int m_star, 
				   int iteration_index, 
				   int new_block_state_no, 
				   int new_block_sector_no, 
				   sector_properties * new_block_sectors_properties_array, 
				   quantum_operator & new_block_static_operator){
  
/*************************************************************************
input variables:  two
    m_star          - the iteration at which the writing process begins. 
    iteration_index - the index of the current iteration. This variable is 
                      written in the file name. 
    new_block_state_no    - the number of new states. 
    new_block_sector_no   - the number of sectors
    new_block_sectors_properties_array - array with the  properties of the given 
                                   sectors. 

output variables: one
    new_block_static_operator - the corresponding operator readed from the file. 
                            the corresponding sector matrices of this operator
                            were obtained in the course of forward NRG 
                            iteration. 

return: bool

What it does:
     This routine reads  one static operator  at a given iteration from the 
     corresponding
     file. this file was created previosuly in the course of the NRG 
     iteration. All the data in the file is readed at once. First the states 
     has to be read in, 
     and the sectors have to be set up before reading the transformation_sectors. 
     The name of the file contains the name of the static_operator and the 
     iteration index. 

How to call:
*************************************************************************/
  
  if (m_star ==-1){
    cout<<"Error in routine read_static_operator(), m_star=-1\nEXITING...\n"; 
  }//if m_star

  
  if((iteration_index >= m_star) && (m_star >= 0)){

    new_block_static_operator.clear_sectors_matrix();
    new_block_static_operator.allocate_sectors_matrix(new_block_sector_no); 


    for(int sector_row=0; sector_row<new_block_sector_no; sector_row++){
      for(int sector_column=0; sector_column<new_block_sector_no; sector_column++){
	
	
	new_block_static_operator.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
										      new_block_sectors_properties_array[sector_column].dimension);


	if(grand_selection_rule(symmetry_no, 
				symmetry_types, 
				new_block_sectors_properties_array[sector_column].representation_index,
				new_block_static_operator.get_representation_index(), 
				new_block_sectors_properties_array[sector_row].representation_index)){
	  
	  

	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].allocate_matrix();
	  
	}//if grand_selection_rule
	  
      }//for sector_column
    }//for sector_row



      

    static_operator_binary_swap_file_name( static_operators_folder_name, 
					     iteration_index,
					     static_operator_file_name,
					     new_block_static_operator
					     );  
    
    static_operator_file_in.open(static_operator_file_name, ios::in|ios::binary); 
    
    if (!static_operator_file_in.good()){
      error_message("reading static_operator_file_in"); 
    }//if

    
    int active_sector_no;

    static_operator_file_in.read(( char *) &active_sector_no, sizeof(int));

    for(int active_sector_index=0; active_sector_index<active_sector_no; active_sector_index++){

      int sector_row; 
      int sector_column; 
      int row_size;
      int column_size; 

      static_operator_file_in.read(( char *) &sector_row, sizeof(int));
      static_operator_file_in.read(( char *) &sector_column, sizeof(int));;
	
      static_operator_file_in.read(( char *) &row_size, sizeof(int)); 
      static_operator_file_in.read(( char *) &column_size, sizeof(int)); 
	
      assert (new_block_static_operator.get_sector_matrix()[sector_row][sector_column].get_height() == row_size); 
      assert (new_block_static_operator.get_sector_matrix()[sector_row][sector_column].get_width() == column_size); 

	
      double matrix_element; 
      for (int i=0; i<row_size; i++){
	for(int j=0;j<column_size; j++){
	  static_operator_file_in.read(( char *) &matrix_element, sizeof(double)); 
	  new_block_static_operator.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i, j, matrix_element);
	}//j
      }//i
	
    }//for active_sector_index

    static_operator_file_in.close();
    static_operator_file_in.clear();

  }//if iteration_index>=m_star
  
  return;
  
}//read_static_operator_binary;








/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void construct_reduced_density_matrix(int symmetry_no,
				      symmetry * symmetry_types,

				      int new_block_sector_no,
				      sector_properties * new_block_sectors_properties_array,
				      int new_block_state_no,
				      state * new_block_states_array,

				      int block_sector_no,
				      sector_properties *block_sectors_properties_array,
				      int block_state_no,
				      state * block_states_array,

				      quantum_operator & new_block_density_matrix,
				      quantum_operator & block_density_matrix,

				      double temperature){

/*************************************************************************
input variables:  two
     symmetry_no   - the number of symmetries in tre present run. 
     symmetry_type - array with the symmetries use

     new_block_sector_no              - the number of sectors in the diagonal basis.
     new_block_sectors_properties     - array with the properties of the sectors in 
                                  the diagonal basis. 
     new_block_state_no               - the number of new states. 
     new_block_states_array           - array with the states in the diagonal basis. 

     block_sector_no          - the number of sectors among the kept states only 
     block_sectors_properties - array with the properties of the sectors in 
                                  current basis (non-orthogonal). 
     block_state_no           - the number of current states. 
     block_states_array       - array with the current states. 

     new_block_density_matrix         - the density matrix in the orthogonal basis 
                                  that has to be reduced by 
                                  integrating the environment  degrees of freedom. 

output variables: one
     block_density_matrix - the reduced density matrix in the kept-kept 
                            channel only obtained from the new_block_density_matrix. 
                            this will be used to construct the new_block_density_matrix of the 
                            previous iteration later.  

return: void

What it does:
     This routine construct the density matrix in the current basis, practically
     construct the kept-kept density matrix in the non-diagonal basis of the
     current states. The discarded-discarded in constructed manually because the matrix
     in this channel is diagonal. The new states_array, new_block_sectors properties as well 
     as the block_states_array and the block_sectors_properties has to be read first. 

How to call:
*************************************************************************/

  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
    if(new_block_density_matrix.get_representation_index(symmetry_index) != 0){
      error_message("checking the quantum numbers of the new_block_density_matrix.\nError in routine create_new_block_hamiltonian\nExiting...\n");
    }//if
  }//for symmetry_index

  

  block_density_matrix.activate(symmetry_no); 
  block_density_matrix.set_name("block_density_matrix"); 
  block_density_matrix.allocate_sectors_matrix(block_sector_no); 

  //we set the representation indices of the new_block_density_matrix to be zero's. 

  for(int i = 0; i < symmetry_no; i++){
    block_density_matrix.set_representation_index(i,0); 
  }//i

  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
    if(block_density_matrix.get_representation_index(symmetry_index) != 0){
      error_message("checking the quantum numbers of the block_density_matrix.\nError in routine construct_reduced_density_matrix\nExiting...\n");
    }//if
  }//for symmetry_index



  for(int sector_row = 0; sector_row < block_sector_no; sector_row++){
    for(int sector_column = 0; sector_column <block_sector_no; sector_column++){
            
      block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_size(block_sectors_properties_array[sector_row].dimension,
										 block_sectors_properties_array[sector_column].dimension);

      
      if(grand_selection_rule(symmetry_no, 
			      symmetry_types, 
			      block_sectors_properties_array[sector_column].representation_index,
			      block_density_matrix.get_representation_index(), 
			      block_sectors_properties_array[sector_row].representation_index)){
	

	block_density_matrix.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, block_sectors_properties_array[sector_row].representation_index);
	block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, block_sectors_properties_array[sector_column].representation_index);
	block_density_matrix.get_sector_matrix()[sector_row][sector_column].allocate_matrix();
      }//if grand_selection_rule      
    }//for sector_column;
  }//for sector_row
    

  /* Now we constrtuct from the new density matrix (in the non-diagonal basis) the block density matrix, i.e.,
     the density matrix in the (kept,kept) sector of the previous iteration (but there in the orthogonal basis) */ 
  
  /* we go through every matrix element of the new density matrix. These give a contribution to the block density 
     matrix of the previous iteration for the states from which this new block state has been constructed  */  

  for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++){
    for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++){

      if(new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].get_activation_state()){

	int height = new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].get_height();
	int width = new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].get_width();

	for(int i = 0; i < height; i++){
	  for(int j = 0; j < width; j++){

	    int state_row = new_block_sectors_properties_array[sector_row].first_state + i;
	    int state_column = new_block_sectors_properties_array[sector_column].first_state + j;
	    
	    /* we look at the states from which this particular new block state has been constructed */

	    int local_state_row = new_block_states_array[state_row].get_local_state_index();
	    int local_state_column = new_block_states_array[state_column].get_local_state_index();
	    
	    int parent_state_row = new_block_states_array[state_row].get_parent_state_index(); 
	    int parent_state_column = new_block_states_array[state_column].get_parent_state_index();
            /* these labels go from 0 to the (no_of_kept_states -1) */

	    if(
	       (local_state_row == local_state_column) 
	       && (block_states_array[parent_state_row].get_representation_index() == block_states_array[parent_state_column].get_representation_index())
	       && (new_block_states_array[state_row].get_representation_index() == new_block_states_array[state_column].get_representation_index())
	       ){
	      
		/* if all quantum numbers match then I compute the degeneracy factor in the recursion formula and
		   add it to the density matrix  */
		int parent_sector_row = block_states_array[parent_state_row].get_sector();
		int parent_sector_column = block_states_array[parent_state_column].get_sector();
		
		int parent_sector_index_row = block_states_array[parent_state_row].get_sector_index();
		int parent_sector_index_column = block_states_array[parent_state_column].get_sector_index();


		double degeneracy_factor =((double)(new_block_sectors_properties_array[sector_row].degeneracy)) 
		    / ((double)(block_sectors_properties_array[parent_sector_row].degeneracy)); 

	      double matrix_element (0);
	      
	      
	      matrix_element = degeneracy_factor * new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j];
	      

	      
	      matrix_element += block_density_matrix.get_sector_matrix()[parent_sector_row][parent_sector_column].
		get_matrix()[parent_sector_index_row][parent_sector_index_column];

	      
	      block_density_matrix.get_sector_matrix()[parent_sector_row][parent_sector_column].
		set_matrix_element(parent_sector_index_row, parent_sector_index_column, matrix_element);
	      
	    }//if local_state_row
	  }//for j
	}//for i
      }//if new_block_density_matrix.get_sector_matrix()      
    }//sector_column
  }//sector_row
  return;   
}//construct_reduced_density_matrix


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


double calculate_trace(int sector_no,
		       sector_properties *sectors_properties_array,		       
		       quantum_operator & one_operator){
  
/*********************************************************************
input variables:  two
     sector_no                - the number of sectors
     sectors_properties_array - the properties of the input sectors. 
     one_operator             - the operator whose trace want to be computed.

output variables: one

return: double that is the value of trace. 


What it does:
    This routine computes the trance of a given operator, and is used mainly for 
    the trace calculation of the reduced density matrix to check that Tr{\ro} = 1.

How to call:
*********************************************************************/

  double trace (0.0); 
  
  for(int sector_row = 0; sector_row < sector_no; sector_row++){
    for(int sector_column = 0; sector_column <sector_no; sector_column++){

      if(one_operator.get_sector_matrix()[sector_row][sector_column].get_activation_state()){

	int height = one_operator.get_sector_matrix()[sector_row][sector_column].get_height();
	int width = one_operator.get_sector_matrix()[sector_row][sector_column].get_width();
	//cout<<"sector_row="<<sector_row<<endl;
	//cout<<"sector_column="<<sector_column<<endl;
	//cout<<"sector_dimension="<<sectors_properties_array[sector_row].dimension<<endl; 
	for(int i = 0; i < height; i++){
	  //cout<<i<<"\t"
	  for(int j = 0; j < width; j++){
	    //cout<< one_operator.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j]<<"\t";
	    if (i==j){
	      trace += sectors_properties_array[sector_row].degeneracy * one_operator.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j];
	    }//if	    
	  }//j
	}//i
	//cout<<trace; 
      }//if one_operator.get_sector_matrix()                 
    }//for sector_column;
  }//for sector_row
  
  return trace; 

}//calculate trace
  


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/




void transfer_reduced_density_matrix(int iteration_index,
				     int symmetry_no,
				     symmetry * symmetry_types,

				     int block_sector_no,
				     sector_properties *block_sectors_properties_array,
				     int block_state_no,
				     state *  block_states_array,				     

				     int * block_original_sector_array, 
				     int * block_original_sector_index_arrray,
				     
				     int new_block_sector_no,
				     sector_properties * new_block_sectors_properties_array,
				     int new_block_state_no,
				     state *  new_block_states_array,
				     
				     double *new_block_energies_array,
				     double *sorted_new_block_energies_array,
				     unsigned long  *new_block_energies_index,
				     
				     quantum_operator & block_density_matrix,				     
				     quantum_operator & new_block_density_matrix,

				     double temperature,
				     double partition_function){

/*************************************************************************
input variables:  12
     symmetry_no                - the number of symmetries in tre present run. 
     symmetry_type              - array with the symmetries use

     block_sector_no          - the number of sectors in the current basis.
     block_sectors_properties - array with the properties of the sectors in 
                                  current basis (non-orthogonal). 
     block_state_no           - the number of current states. 
     block_states_array       - array with the current states. 


     block_original_sector_array    - original sectors in new_block_states from which block_states were choped off
     block_original_sector_index_array  - original (before the elimination process) indices of kept states (i.e block_states) within 
                                          the sectors in new_block_states from which the other block_states have been eliminated 
                                          

     note that these new_block_... are the new_block_... parameters at the previous iteration. 

     new_block_sector_no              - the number of sectors in the diagonal basis.
     new_block_sectors_properties     - array with the properties of the sectors in the 
                                  diagonal basis. 
     new_block_state_no               - the number of new states. 
     new_block_states_array           - array with the states in the diagonal basis. 

     block_density_matrix - the reduced density matrix in the kept-kept channel


output variables: one
    new_block_density_matrix - the density matrix at the previous iteration. The 
                         kept-kept channel is tranfered to the 
                         new_block_density_matrix while the discarded-discarded 
                         channel is constructed by hand. for the case of T=0 
                         the discarded-discarded channel has only zero matrix 
                         elements. 
      
return: bool

What it does:
     This routine construct the density matrix in the orthogonal basis, by combining
     the kept-kept and the discarded-discarded channels. the kept-discarded
     and discarded-kept sectors has only zero matrix elements. This corresponds to 
     the unreduced density matrix at the previous iteration. 

How to call:
*************************************************************************/

  //first we clear the new block density matrix. 



    for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
	if(block_density_matrix.get_representation_index(symmetry_index) != 0){
	    error_message("checking the quantum numbers of the block_density_matrix.\nError in routine transfer_reduced_density_matrix\nExiting...\n");
	}//if
    }//for symmetry_index
    
    new_block_density_matrix.activate(symmetry_no); 
    new_block_density_matrix.set_name("new_block_density_matrix"); 
    new_block_density_matrix.allocate_sectors_matrix(new_block_sector_no); 

    //we set the representation indices of the new_block_density_matrix to be zero's. 

    for(int i = 0; i < symmetry_no; i++){
	new_block_density_matrix.set_representation_index(i,0); 
    }//i

    for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
	if(new_block_density_matrix.get_representation_index(symmetry_index) != 0){
	    error_message("checking the quantum numbers of the new_block_density_matrix.\nError in routine transfer_reduced_density_matrix\nExiting...\n");
	}//if
    }//for symmetry_index
  
    
    for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++){
	for(int sector_column = 0; sector_column <new_block_sector_no; sector_column++){
            
	    new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
											     new_block_sectors_properties_array[sector_column].dimension);
	    
	    
	    if(grand_selection_rule(symmetry_no, 
				    symmetry_types, 
				    new_block_sectors_properties_array[sector_column].representation_index,
				    new_block_density_matrix.get_representation_index(), 
				    new_block_sectors_properties_array[sector_row].representation_index)){
		
		
		new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
		new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
		new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
		new_block_density_matrix.get_sector_matrix()[sector_row][sector_column].allocate_matrix();
		
	    }//if      
	}//for sector_column
    }//for sector_row



  for(int block_sector_row = 0; block_sector_row < block_sector_no; block_sector_row++){
	    for(int block_sector_column = 0; block_sector_column < block_sector_no; block_sector_column++){
	      if(block_density_matrix.get_sector_matrix()[block_sector_row][block_sector_column].get_activation_state()){

		int block_height = block_density_matrix.get_sector_matrix()[block_sector_row][block_sector_column].get_height();
		int block_width = block_density_matrix.get_sector_matrix()[block_sector_row][block_sector_column].get_width();

		for (int block_i=0; block_i < block_height; block_i++){
		  for (int block_j=0; block_j < block_width; block_j++){
		    
		    int block_state_row= block_sectors_properties_array[block_sector_row].first_state+block_i; 
		    int block_state_column = block_sectors_properties_array[block_sector_column].first_state+block_j; 

		    double matrix_element = block_density_matrix.get_sector_matrix()[block_sector_row][block_sector_column].
		      get_matrix()[block_i][block_j]; 			    		   

		    new_block_density_matrix.get_sector_matrix()[block_original_sector_array[block_state_row]][block_original_sector_array[block_state_column]].set_matrix_element(block_original_sector_index_array[block_state_row], block_original_sector_index_array[block_state_column], matrix_element); 
		    
		  }//for block_j
		}//for block_i

	      }//if block_density_matrix.get_sector_matrix
	    }//for block_sector_column
  }//for block_sector_row


  //this is the discarded-discarded channel for finite temperatures


  if (temperature> 0){

    
    double energy_min = sorted_new_block_energies_array[0]; 
    double rescaled_temperature = temperature*pow(lambda, 0.5*((double)(iteration_index)));
    double local_degeneracy = get_local_degeneracy(symmetry_no, 
						   symmetry_types, 
						   
						   new_block_sector_no, 
						   new_block_sectors_properties_array
						   ); 
    

    for(int new_block_sector = 0; new_block_sector < new_block_sector_no; new_block_sector++){  
      if(new_block_density_matrix.get_sector_matrix()[new_block_sector][new_block_sector].get_activation_state()){	
	int new_block_size = new_block_density_matrix.get_sector_matrix()[new_block_sector][new_block_sector].get_height();
	for (int i=0; i< new_block_size; i++){	
	  int new_block_state = new_block_sectors_properties_array[new_block_sector].first_state+i; 	  

	  /* If the state has been eliminated in the forward iteration step then we set its weight in the 
	     reduced density matrix by hand. */
	  if (new_block_states_array[new_block_state].get_elimination_tag()=='D'){
	    
	    double matrix_element = (1.0/partition_function)
	      *exp(-(new_block_energies_array[new_block_state]-energy_min)/rescaled_temperature)
	      *pow(local_degeneracy,((double)(iteration_no-iteration_index)));
	    
	    new_block_density_matrix.get_sector_matrix()[new_block_sector][new_block_sector].set_matrix_element(i, i, matrix_element); 
	  }//if new_block_states_array
	}//i
	
      }//if new_block_density_matrix   
    }//for new_block_sector
  }//if temperature;
  

  return; 


}//transfer_reduced_density_matrix



