(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[     98555,       2434]
NotebookOptionsPosition[     96073,       2359]
NotebookOutlinePosition[     96427,       2375]
CellTagsIndexPosition[     96384,       2372]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[TextData[{
 "Kondo model  with ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    SubscriptBox["SU", "charge"],
    FontSlant->"Italic"], TraditionalForm]]],
 "(2) \[Cross] ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    SubscriptBox["SU", "spin"],
    FontSlant->"Italic"], TraditionalForm]]],
 "(2) symmetries"
}], "Title",
 CellChangeTimes->{{3.422598484837209*^9, 3.42259849486808*^9}, {
  3.422599040637944*^9, 3.422599046758636*^9}}],

Cell["Initialization ", "Subtitle"],

Cell[BoxData[{
 RowBox[{"ClearAll", "[", "\"\<Global`*\>\"", "]"}], "\[IndentingNewLine]", 
 RowBox[{"Off", "[", 
  RowBox[{
   RowBox[{"General", "::", "spell"}], ",", " ", 
   RowBox[{"General", "::", "spell1"}]}], "]"}]}], "Input"],

Cell[TextData[{
 "This notebook describes how to construct the input matrices for the case \
when two symmetries are considered U(1) for the charge and U(1) for the spin, \
corresponding to the total charge with respect to half-filling and the  \
z-component of the total spin, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", 
     RowBox[{"z", ",", "total"}]], ".", " "}], TraditionalForm]]],
 "\nWe shall represent the spin using fermions. The total charge is computed \
as \n\t",
 Cell[BoxData[
  FormBox[
   SubscriptBox["Q", "N"], TraditionalForm]]],
 " = ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     FormBox[
      SuperscriptBox[
       SubscriptBox["\[Sum]", 
        RowBox[{"n", "=", "0"}]], "N"],
      TraditionalForm], 
     RowBox[{
      SuperscriptBox[
       SubscriptBox["f", "n"], "+"], 
      SubscriptBox["f", "n"]}]}], " ", "+", 
    RowBox[{
     SuperscriptBox["d", "+"], "d"}], " ", "-", " ", 
    RowBox[{"(", 
     RowBox[{"N", "+", "1"}], ")"}]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["f", "n"], TraditionalForm]]],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    SuperscriptBox[
     SubscriptBox["f", "n"], "+"]}], TraditionalForm]]],
 ") is the annihilation (creation) operator for the band electrons and ",
 StyleBox["d",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", " ", 
    SuperscriptBox["d", "+"], ")"}], TraditionalForm]]],
 " is the annihilation (creation) operator for the d-levels. The local spin  \
will be represented through the fermion ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["d", "+"], TraditionalForm]]],
 " . In the 0-th iteration,  we have thus two electronic (single particle) \
levels with two spins,  i.e.,  FOUR  single particle states:  \n\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox[
     SubscriptBox["f", "\[UpArrow]"], "+"], "|", 
    RowBox[{"0", ">"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox[
      SubscriptBox["f", "\[DownArrow]"], "+"], "|", 
     RowBox[{"0", ">"}]}], ",", " ", 
    RowBox[{
     SuperscriptBox[
      SubscriptBox["d", "\[UpArrow]"], "+"], "|", 
     RowBox[{"0", ">"}]}], ",", 
    FormBox[
     RowBox[{
      SuperscriptBox[
       SubscriptBox["d", "\[DownArrow]"], "+"], "|", 
      RowBox[{"0", ">", "\[IndentingNewLine]"}]}],
     TraditionalForm]}], TraditionalForm]]],
 "\nwith  |0> the vacuum state than the single-electron states are. We shall \
therefore first construct a ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["2", "4"], "=", "16", " "}], TraditionalForm]]],
 "-dimensional  Hilbert spapce, generated by the operators ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["f", "\[UpArrow]"], "+"], TraditionalForm]]],
 ",  ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["f", "\[DownArrow]"], "+"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["d", "\[UpArrow]"], "+"], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["d", "\[DownArrow]"], "+"], TraditionalForm]]],
 ", and define fermionic creation and annihilation operators there. These \
operators shall then be used to construct symbolically the many-body states \
and operators. \nIn the Kondo model, the constraint is that the d-level is \
allways  singly-occupied, so after constructing the states we will keep only \
states with occupation one for the d-level. "
}], "Text",
 CellChangeTimes->{{3.422257807746289*^9, 3.422258008894191*^9}, {
  3.422258058754923*^9, 3.422258072175811*^9}, {3.422258129598818*^9, 
  3.422258130575411*^9}, {3.422259264756965*^9, 3.422259265264323*^9}, {
  3.422259320850274*^9, 3.422259651493224*^9}, {3.422259689880664*^9, 
  3.422259703369674*^9}, {3.422259772897277*^9, 3.422259784098445*^9}, {
  3.422262329515971*^9, 3.422262518108431*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfSingleOccupiedStates", "=", "4"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"NumberOfTotalStates", " ", "=", " ", 
   SuperscriptBox["2", "NumberOfSingleOccupiedStates"]}], ";"}]}], "Input",
 CellChangeTimes->{{3.42225970942581*^9, 3.422259753959048*^9}, {
  3.422260356857766*^9, 3.422260359126533*^9}}],

Cell["\<\
Next we construct effectively the NumberOfTotalStates * NumberOfTotalStates \
-dimensional creation operator matrices. This part you can skip. \
\>", "Text",
 CellChangeTimes->{{3.422260232752746*^9, 3.422260239500793*^9}, {
  3.422260278460007*^9, 3.422260287823447*^9}, {3.422260459641082*^9, 
  3.422260469284508*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Occupation", " ", "=", " ", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"IntegerDigits", "[", 
      RowBox[{"i", ",", "2", ",", " ", "NumberOfSingleOccupiedStates"}], 
      "]"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"i", ",", "0", ",", 
       RowBox[{"NumberOfTotalStates", "-", "1"}]}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Operator", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfSingleOccupiedStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"OperatorDagger", " ", "=", "Operator"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{"For", "[", 
  RowBox[{
   RowBox[{"r", "=", "1"}], ",", " ", 
   RowBox[{"r", "\[LessEqual]", " ", "NumberOfSingleOccupiedStates"}], ",", 
   " ", 
   RowBox[{"r", "++"}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"For", "[", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", " ", 
      RowBox[{"i", "\[LessEqual]", "NumberOfTotalStates"}], ",", " ", 
      RowBox[{"i", "++"}], ",", "\[IndentingNewLine]", 
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", " ", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfTotalStates"}], ",", " ", 
        RowBox[{"j", "++"}], ",", "\[IndentingNewLine]", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"For", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"state", "=", "1"}], ";", " ", 
            RowBox[{"StateSign", "=", "1"}]}], ",", " ", 
           RowBox[{"state", "<", "r"}], ",", " ", 
           RowBox[{"state", "++"}], ",", "\[IndentingNewLine]", 
           RowBox[{"StateSign", "*=", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"-", "1"}], ")"}], 
             RowBox[{"Occupation", "\[LeftDoubleBracket]", 
              RowBox[{"j", ",", " ", "state"}], 
              "\[RightDoubleBracket]"}]]}]}], "]"}], ";", 
         "\[IndentingNewLine]", "\[IndentingNewLine]", 
         RowBox[{"If", "[", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"FromDigits", "[", 
               RowBox[{
                RowBox[{"BitXor", "[", 
                 RowBox[{
                  RowBox[{
                  "Occupation", "\[LeftDoubleBracket]", "i", 
                   "\[RightDoubleBracket]"}], ",", " ", 
                  RowBox[{
                  "Occupation", "\[LeftDoubleBracket]", "j", 
                   "\[RightDoubleBracket]"}]}], "]"}], ",", "2"}], "]"}], " ",
               "\[Equal]", " ", 
              SuperscriptBox["2", 
               RowBox[{"NumberOfSingleOccupiedStates", "-", "r"}]]}], ")"}], 
            "\[IndentingNewLine]", "&&", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"FromDigits", "[", 
               RowBox[{
               "Occupation", "\[LeftDoubleBracket]", "i", 
                "\[RightDoubleBracket]"}], "]"}], ">", 
              RowBox[{"FromDigits", "[", 
               RowBox[{
               "Occupation", "\[LeftDoubleBracket]", "j", 
                "\[RightDoubleBracket]"}], "]"}]}], ")"}]}], ",", 
           "\[IndentingNewLine]", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"OperatorDagger", "\[LeftDoubleBracket]", 
             RowBox[{"r", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], 
            "=", "StateSign"}]}], "\[IndentingNewLine]", 
          "\[IndentingNewLine]", "]"}]}]}], "\[IndentingNewLine]", "]"}]}], 
     "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
     "Operator", "\[LeftDoubleBracket]", "r", "\[RightDoubleBracket]"}], " ", 
     "=", 
     RowBox[{"Transpose", "[", 
      RowBox[{
      "OperatorDagger", "\[LeftDoubleBracket]", "r", 
       "\[RightDoubleBracket]"}], "]"}]}]}]}], "\[IndentingNewLine]", 
  "]"}]}], "Input",
 CellChangeTimes->{{3.42226025055771*^9, 3.422260260564144*^9}}],

Cell["\<\
We have now     NumberOfSingleOccupiedStates = 4 creation operators:  \
OperatorDagger\[LeftDoubleBracket]1\[RightDoubleBracket],..., OperatorDagger\
\[LeftDoubleBracket]4\[RightDoubleBracket]. These shall be used to construct \
the necessary operators: \
\>", "Text",
 CellChangeTimes->{{3.422257807746289*^9, 3.422258008894191*^9}, {
   3.422258058754923*^9, 3.422258072175811*^9}, {3.422258129598818*^9, 
   3.422258130575411*^9}, {3.422259264756965*^9, 3.422259265264323*^9}, {
   3.422259320850274*^9, 3.422259651493224*^9}, {3.422259689880664*^9, 
   3.422259703369674*^9}, {3.422259772897277*^9, 3.422259784098445*^9}, {
   3.422260330982351*^9, 3.422260334508321*^9}, {3.422260366221652*^9, 
   3.422260446806384*^9}, 3.422260485771486*^9}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"fDaggerUp", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fDaggerDown", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDaggerUp", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDaggerDown", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fUp", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fDown", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dUp", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDown", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Szd", "=", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"dDaggerUp", ".", "dUp"}], "-", 
      RowBox[{"dDaggerDown", ".", "dDown"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SPlusd", " ", "=", 
   RowBox[{"dDaggerUp", ".", "dDown"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SMinusd", " ", "=", 
   RowBox[{"dDaggerDown", ".", "dUp"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Szf", "=", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"fDaggerUp", ".", "fUp"}], "-", 
      RowBox[{"fDaggerDown", ".", "fDown"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SPlusf", " ", "=", 
   RowBox[{"fDaggerUp", ".", "fDown"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SMinusf", " ", "=", 
   RowBox[{"fDaggerDown", ".", "fUp"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Sz", "=", 
   RowBox[{"Szd", "+", "Szf"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SPlus", "=", 
   RowBox[{"SPlusd", "+", "SPlusf"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SMinus", "=", 
    RowBox[{"SMinusd", "+", "SMinusf"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"UnitOperator", " ", "=", " ", 
    RowBox[{
     RowBox[{"dDaggerUp", ".", "dUp"}], "+", 
     RowBox[{"dUp", ".", "dDaggerUp"}]}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Qzd", "=", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"dDaggerUp", ".", "dUp"}], " ", "+", 
      RowBox[{"dDaggerDown", ".", "dDown"}], " ", "-", " ", "UnitOperator"}], 
     ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"QPlusd", " ", "=", 
   RowBox[{"dDaggerUp", ".", "dDown"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"QMinusd", " ", "=", 
   RowBox[{"dDaggerDown", ".", "dUp"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Qzf", "=", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"fDaggerUp", ".", "fUp"}], " ", "+", 
      RowBox[{"fDaggerDown", ".", "fDown"}], " ", "-", " ", "UnitOperator"}], 
     ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"QPlusf", " ", "=", 
   RowBox[{"fDaggerUp", ".", "fDown"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"QMinusf", " ", "=", 
   RowBox[{"fDaggerDown", ".", "fUp"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Qz", "=", 
   RowBox[{"Qzd", "+", "Qzf"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"QPlus", "=", 
   RowBox[{"QPlusd", "+", "QPlusf"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"QMinus", "=", 
   RowBox[{"QMinusd", "+", "QMinusf"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.422273235977622*^9, 3.422273314174119*^9}, {
  3.422513221796632*^9, 3.422513250741941*^9}, {3.422599224107562*^9, 
  3.42259935815657*^9}}],

Cell["Now we can construct all the states. ", "Text",
 CellChangeTimes->{{3.422260563311906*^9, 3.422260571250809*^9}}],

Cell["States", "Subtitle"],

Cell["\<\
In this section we construct the corresponding states and order them as in \
the representation index table. First we construct the VacuumState and then \
based on it, we construct all the other states that are allowed. We can \
construct based on these states triplet as well as singlet states. These \
results can be extended to the case when SU(2) symmetries for charge or spin \
are considered. We could also use other basis states and mix State4 and \
State5 \
\>", "Text",
 CellChangeTimes->{{3.422260600765262*^9, 3.422260689774695*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"VacuumState", " ", "=", "  ", 
   RowBox[{"Table", "[", 
    RowBox[{"IntegerDigits", "[", 
     RowBox[{
      SuperscriptBox["2", 
       RowBox[{"NumberOfTotalStates", "-", "1"}]], ",", "2", ",", " ", 
      "NumberOfTotalStates"}], "]"}], "]"}]}], ";"}]], "Input"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfBlockStates", " ", "=", " ", "3"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"BlockStates", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State1", " ", "=", " ", 
    RowBox[{
     FractionBox["1", 
      SqrtBox["2"]], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"fDaggerDown", ".", "dDaggerUp"}], "-", 
        RowBox[{"fDaggerUp", ".", "dDaggerDown"}]}], ")"}], ".", 
      "VacuumState"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State2", "=", " ", 
    RowBox[{"fDaggerUp", ".", "dDaggerUp", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State3", "=", " ", 
    RowBox[{
    "fDaggerUp", ".", "fDaggerDown", ".", "dDaggerUp", ".", 
     "VacuumState"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.42226084656336*^9, 3.422260942028706*^9}, {
   3.422261115279298*^9, 3.422261172794231*^9}, {3.422598511661529*^9, 
   3.422598585681973*^9}, {3.422598623454855*^9, 3.422598637049208*^9}, {
   3.422599095158405*^9, 3.422599115588187*^9}, {3.422600910011628*^9, 
   3.422600911730426*^9}, {3.422601094971406*^9, 3.422601111684906*^9}, {
   3.422601217344871*^9, 3.42260122665104*^9}, {3.42260129970921*^9, 
   3.42260130978965*^9}, {3.422603066445888*^9, 3.4226030728883*^9}, 
   3.422603105598201*^9}],

Cell["\<\
Now we have to check that these states are indeed ordered.  If not, we have \
to reshuffle them.\
\>", "Text",
 CellChangeTimes->{{3.422260756543946*^9, 3.422260774656064*^9}, {
  3.422261998378347*^9, 3.42226200768368*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfSymmetries", "=", "2"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedQuantumNumbers", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfSymmetries", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]", "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "Qz", ".", 
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "Sz", ".", 
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     "\[IndentingNewLine]", 
     RowBox[{"{", " ", 
      RowBox[{
       RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", " ", 
       RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], " ", "//",
   " ", "MatrixForm"}]}], "Input",
 CellChangeTimes->{{3.422261008583634*^9, 3.422261105170266*^9}, {
   3.422261185877381*^9, 3.422261257246138*^9}, {3.422261290925316*^9, 
   3.42226138106269*^9}, {3.422261556601073*^9, 3.422261621932931*^9}, {
   3.422261667725704*^9, 3.422261669214305*^9}, {3.422261754271065*^9, 
   3.422261761177845*^9}, {3.422261827807973*^9, 3.422261982154367*^9}, {
   3.422273329630683*^9, 3.422273330834617*^9}, {3.422424163530083*^9, 
   3.422424182641464*^9}, {3.422424283395106*^9, 3.422424295014417*^9}, {
   3.422424504284488*^9, 3.422424529308189*^9}, {3.422425828934438*^9, 
   3.422425948432719*^9}, {3.422426022264904*^9, 3.422426031542518*^9}, {
   3.422426385786546*^9, 3.422426386364774*^9}, {3.42251326110266*^9, 
   3.422513266106889*^9}, 3.422598651038154*^9, {3.422599392205433*^9, 
   3.422599399846498*^9}, {3.4226009462946978`*^9, 3.422600979477351*^9}, {
   3.422601761063846*^9, 3.422601765106821*^9}}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", "0"},
     {"0", "1"},
     {
      FractionBox["1", "2"], 
      FractionBox["1", "2"]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{{3.42226107342192*^9, 3.422261105951935*^9}, 
   3.422261189857225*^9, {3.422261361886205*^9, 3.422261382372428*^9}, {
   3.422261557654303*^9, 3.422261582070163*^9}, 3.422261670042143*^9, 
   3.4222618584621*^9, {3.422261936692001*^9, 3.422261947147733*^9}, 
   3.422261982631925*^9, 3.422273333398469*^9, 3.422422421961272*^9, 
   3.422424185206617*^9, {3.422424284343575*^9, 3.422424308315455*^9}, {
   3.422424505813002*^9, 3.42242452992562*^9}, 3.422425952720056*^9, {
   3.422426001119572*^9, 3.422426038231274*^9}, 3.422426387262979*^9, 
   3.422427033012301*^9, 3.422427279541733*^9, 3.42243318197125*^9, 
   3.42251304385896*^9, 3.422513278159466*^9, 3.422513834119571*^9, 
   3.4225932132783*^9, 3.422598666855304*^9, 3.422599118727433*^9, 
   3.422599400798146*^9, 3.42260092178914*^9, 3.422600980052557*^9, 
   3.422601063926791*^9, {3.422601101571587*^9, 3.422601116208975*^9}, {
   3.422601220437288*^9, 3.42260122870376*^9}, {3.422601302017582*^9, 
   3.422601314278729*^9}, 3.422601766095554*^9, 3.422602827293759*^9, 
   3.42260307539983*^9, 3.422603115097525*^9, 3.422641141034209*^9, 
   3.422685769099748*^9, 3.422696159974625*^9, 3.42286440575963*^9}]
}, Open  ]],

Cell["\<\
We now calculate the reduced matrix elements of the hopping and spectral \
operators using the Wigner-Eckart theorem\
\>", "Text",
 CellChangeTimes->{3.422598725708222*^9}],

Cell["Spectral and hopping operators.", "Subtitle"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"OrderedfDaggerUp", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"OrderedfDaggerDown", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"OrderedfUp", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedfDown", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]", "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"OrderedfDaggerUp", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDaggerUp", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedfDaggerDown", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDaggerDown", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedfUp", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fUp", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedfDown", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDown", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SpinReducedfDagger", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SpinReducedf", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Reducedf", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]", "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Table", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"OrderedfDaggerUp", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]",
            "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SpinReducedfDagger", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ",
            
           RowBox[{
            RowBox[{"OrderedfDaggerUp", "\[LeftDoubleBracket]", 
             RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
            RowBox[{"ClebschGordan", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                FractionBox["1", "2"], ",", 
                FractionBox["1", "2"]}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], 
               "}"}]}], "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"OrderedfDaggerDown", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]",
            "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SpinReducedfDagger", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ",
            
           RowBox[{
            RowBox[{"OrderedfDaggerDown", "\[LeftDoubleBracket]", 
             RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
            RowBox[{"ClebschGordan", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                FractionBox["1", "2"], ",", 
                RowBox[{"-", 
                 FractionBox["1", "2"]}]}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], 
               "}"}]}], "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        "\[IndentingNewLine]", 
        RowBox[{"{", 
         RowBox[{"SpinReducedfDagger", "\[LeftDoubleBracket]", 
          RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
       " ", 
       RowBox[{"{", 
        RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
     "MatrixForm"}], "//", "OutputForm"}], "//", "N"}], "\[IndentingNewLine]",
   "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Table", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"OrderedfUp", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]",
            "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SpinReducedf", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ",
            
           RowBox[{
            RowBox[{"-", 
             RowBox[{"OrderedfUp", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}]}], "/", 
            RowBox[{"ClebschGordan", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                FractionBox["1", "2"], ",", 
                RowBox[{"-", 
                 FractionBox["1", "2"]}]}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], 
               "}"}]}], "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"OrderedfDown", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]",
            "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SpinReducedf", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ",
            
           RowBox[{
            RowBox[{"OrderedfDown", "\[LeftDoubleBracket]", 
             RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
            RowBox[{"ClebschGordan", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                FractionBox["1", "2"], ",", 
                FractionBox["1", "2"]}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], 
               "}"}]}], "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        "\[IndentingNewLine]", 
        RowBox[{"{", 
         RowBox[{"SpinReducedf", "\[LeftDoubleBracket]", 
          RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
       " ", 
       RowBox[{"{", 
        RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
     "MatrixForm"}], "//", "OutputForm"}], "//", "N"}], "\[IndentingNewLine]",
   "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"SpinReducedfDagger", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"Reducedf", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"SpinReducedfDagger", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               FractionBox["1", "2"]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"SpinReducedf", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"Reducedf", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"SpinReducedf", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               RowBox[{"-", 
                FractionBox["1", "2"]}]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"Reducedf", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
      " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422427108128412*^9, 3.422427111736916*^9}, {
   3.422427153114619*^9, 3.422427212919719*^9}, 3.422427535099241*^9, {
   3.422433150378739*^9, 3.422433152654368*^9}, 3.422598714687874*^9, {
   3.42259951802725*^9, 3.422599662551419*^9}, {3.422599743773122*^9, 
   3.422599805828801*^9}, {3.422599859175422*^9, 3.422599891244227*^9}, {
   3.422600024103885*^9, 3.422600139259238*^9}, {3.422600170783361*^9, 
   3.422600190387934*^9}, {3.422600253273129*^9, 3.422600272460073*^9}, {
   3.422600305465586*^9, 3.422600516662481*^9}, {3.422600548778142*^9, 
   3.422600615079454*^9}, {3.422600793202635*^9, 3.422600813100365*^9}, {
   3.422600859192306*^9, 3.42260088806534*^9}, {3.422601822808557*^9, 
   3.42260188379778*^9}, {3.422602848985998*^9, 3.422602860182903*^9}, 
   3.422641327365824*^9, {3.422641389002624*^9, 3.42264140149065*^9}}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}}, {{0.7071067811865476}, \
{-1.224744871391589}, {0.}}}\
\>", "\<\
0.         0.         0.

0.         0.         0.

0.707107   -1.22474   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422427009374603*^9, 3.422427041009018*^9, 3.422427112761862*^9, {
   3.422427155075844*^9, 3.422427213797051*^9}, 3.422427281483832*^9, 
   3.422433185296861*^9, 3.422513049729296*^9, 3.422513283827814*^9, 
   3.422513836491431*^9, 3.422593215638943*^9, 3.422598728747492*^9, 
   3.422599448396893*^9, {3.422599553643453*^9, 3.422599565424723*^9}, 
   3.422599667637441*^9, {3.422599746617936*^9, 3.422599806655919*^9}, 
   3.422599878518485*^9, {3.422600044608358*^9, 3.422600103885184*^9}, 
   3.422600139988218*^9, {3.422600176626048*^9, 3.422600191012933*^9}, {
   3.422600483848408*^9, 3.422600527537164*^9}, 3.422600615806196*^9, 
   3.42260081437882*^9, 3.422600861472409*^9, 3.422600891480189*^9, 
   3.422600932405465*^9, {3.422601823543368*^9, 3.422601884599177*^9}, {
   3.422602818659533*^9, 3.422602860770585*^9}, 3.422603082314864*^9, 
   3.422603116992787*^9, 3.422641235854549*^9, 3.422641331546104*^9, {
   3.422641392856023*^9, 3.422641402231889*^9}, 3.422685772895037*^9, 
   3.422696161787518*^9, 3.422864407980257*^9}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {-1.}}, {{0.}, {0.}, {-1.}}, {{0.}, {0.}, {0.}}}\
\>", "\<\
0.    0.    -1.

0.    0.    -1.

0.    0.    0.\
\>"], "Output",
 CellChangeTimes->{
  3.422427009374603*^9, 3.422427041009018*^9, 3.422427112761862*^9, {
   3.422427155075844*^9, 3.422427213797051*^9}, 3.422427281483832*^9, 
   3.422433185296861*^9, 3.422513049729296*^9, 3.422513283827814*^9, 
   3.422513836491431*^9, 3.422593215638943*^9, 3.422598728747492*^9, 
   3.422599448396893*^9, {3.422599553643453*^9, 3.422599565424723*^9}, 
   3.422599667637441*^9, {3.422599746617936*^9, 3.422599806655919*^9}, 
   3.422599878518485*^9, {3.422600044608358*^9, 3.422600103885184*^9}, 
   3.422600139988218*^9, {3.422600176626048*^9, 3.422600191012933*^9}, {
   3.422600483848408*^9, 3.422600527537164*^9}, 3.422600615806196*^9, 
   3.42260081437882*^9, 3.422600861472409*^9, 3.422600891480189*^9, 
   3.422600932405465*^9, {3.422601823543368*^9, 3.422601884599177*^9}, {
   3.422602818659533*^9, 3.422602860770585*^9}, 3.422603082314864*^9, 
   3.422603116992787*^9, 3.422641235854549*^9, 3.422641331546104*^9, {
   3.422641392856023*^9, 3.422641402231889*^9}, 3.422685772895037*^9, 
   3.422696161787518*^9, 3.422864408024227*^9}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {-1.4142135623730951}}, {{0.}, {0.}, {-1.4142135623730951}}, \
{{0.7071067811865476}, {-1.224744871391589}, {0.}}}\
\>", "\<\
0.         0.         -1.41421

0.         0.         -1.41421

0.707107   -1.22474   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422427009374603*^9, 3.422427041009018*^9, 3.422427112761862*^9, {
   3.422427155075844*^9, 3.422427213797051*^9}, 3.422427281483832*^9, 
   3.422433185296861*^9, 3.422513049729296*^9, 3.422513283827814*^9, 
   3.422513836491431*^9, 3.422593215638943*^9, 3.422598728747492*^9, 
   3.422599448396893*^9, {3.422599553643453*^9, 3.422599565424723*^9}, 
   3.422599667637441*^9, {3.422599746617936*^9, 3.422599806655919*^9}, 
   3.422599878518485*^9, {3.422600044608358*^9, 3.422600103885184*^9}, 
   3.422600139988218*^9, {3.422600176626048*^9, 3.422600191012933*^9}, {
   3.422600483848408*^9, 3.422600527537164*^9}, 3.422600615806196*^9, 
   3.42260081437882*^9, 3.422600861472409*^9, 3.422600891480189*^9, 
   3.422600932405465*^9, {3.422601823543368*^9, 3.422601884599177*^9}, {
   3.422602818659533*^9, 3.422602860770585*^9}, 3.422603082314864*^9, 
   3.422603116992787*^9, 3.422641235854549*^9, 3.422641331546104*^9, {
   3.422641392856023*^9, 3.422641402231889*^9}, 3.422685772895037*^9, 
   3.422696161787518*^9, 3.422864408069959*^9}]
}, Open  ]],

Cell["Hamiltonian terms", "Subtitle"],

Cell[TextData[{
 "In this section we consider different terms appearing in the Hamiltonian \
for the Kondo Hamiltonian. The expression used for it : ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["H", "Kondo"], "=", 
    RowBox[{
     FractionBox["1", "2"], "J", 
     RowBox[{"\[Sum]", " ", 
      RowBox[{
       OverscriptBox["S", 
        RowBox[{"\[Rule]", " "}]], " ", 
       SuperscriptBox["f", "+"], "\[Sigma]", " ", "f"}]}]}]}], 
   TraditionalForm]]],
 "  "
}], "Text",
 CellChangeTimes->{{3.42251306960253*^9, 3.422513070535576*^9}, {
  3.42251312212709*^9, 3.422513128759267*^9}, {3.42251316560735*^9, 
  3.422513179294315*^9}, {3.422598763391312*^9, 3.422598768128859*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfHamiltonianTerms", "=", "1"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"HKondo", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfHamiltonianTerms", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"HKondo", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], 
    "=", 
    RowBox[{
     RowBox[{"Szd", ".", "Szf"}], " ", "+", " ", 
     RowBox[{
      FractionBox["1", "2"], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"SPlusd", ".", "SMinusf"}], " ", "+", " ", 
        RowBox[{"SMinusd", ".", "SPlusf"}]}], ")"}]}]}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedHamiltonian", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfHamiltonianTerms", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderedHamiltonian", "\[LeftDoubleBracket]", 
         RowBox[{"k", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], " ", 
        "=", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", 
         RowBox[{
         "HKondo", "\[LeftDoubleBracket]", "k", "\[RightDoubleBracket]"}], 
         ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"k", ",", "NumberOfHamiltonianTerms"}], "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422427388370775*^9, 3.422427437070901*^9}, {
  3.42242749172612*^9, 3.422427526825983*^9}, {3.422427584982783*^9, 
  3.422427590687387*^9}, {3.422433136386855*^9, 3.422433139040832*^9}, {
  3.422513104824159*^9, 3.422513107084334*^9}, {3.422513182350112*^9, 
  3.422513182441258*^9}, {3.422513298533192*^9, 3.422513434248784*^9}, {
  3.422598782036206*^9, 3.422598785649424*^9}}],

Cell[OutputFormData["\<\
{{{{-0.75}, {0.}, {0.}}, {{0.}, {0.25}, {0.}}, {{0.}, {0.}, {0.}}}}\
\>", "\<\
-0.75   0.      0.
0.      0.25    0.
0.      0.      0.\
\>"], "Output",
 CellChangeTimes->{
  3.422257525369031*^9, 3.422262250568518*^9, 3.42242243161654*^9, {
   3.422427510858111*^9, 3.422427528394162*^9}, 3.422427592710337*^9, 
   3.422433191045149*^9, 3.422513434909226*^9, 3.422513839720494*^9, 
   3.422593227115342*^9, 3.422598788979146*^9, 3.422601913914409*^9, 
   3.422641238925959*^9, 3.42264145431794*^9, 3.422685802376637*^9, 
   3.422696175886692*^9}]
}, Open  ]],

Cell["Spin operator", "Subtitle",
 CellChangeTimes->{{3.422257543929043*^9, 3.422257549617286*^9}}],

Cell["In this section we consider the spin operator", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422513148969978*^9, 3.422513151758013*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"OrderedSPlus", " ", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"OrderedSz", " ", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedSMinus", " ", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"OrderedSPlus", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{"-", 
         FractionBox["1", 
          SqrtBox["2"]]}], " ", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", "SPlusd", ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedSz", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "Szd", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedSMinus", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        FractionBox["1", 
         SqrtBox["2"]], " ", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", "SMinusd", ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ReducedS", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]", "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrderedSPlus", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedS", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrderedSPlus", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{"1", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrderedSz", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedS", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrderedSz", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{"1", ",", "0"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrderedSMinus", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedS", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrderedSMinus", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{"1", ",", 
               RowBox[{"-", "1"}]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"ReducedS", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
      " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422257650153471*^9, 3.422257658110106*^9}, {
   3.422262270013571*^9, 3.422262273937019*^9}, {3.422427554252507*^9, 
   3.422427571961936*^9}, {3.42242760434566*^9, 3.422427621897156*^9}, {
   3.422433129489568*^9, 3.422433132897229*^9}, 3.422513455801693*^9, 
   3.422641262720333*^9}],

Cell[OutputFormData["\<\
{{{0.}, {-0.8660254037844386}, {0.}}, {{0.5}, {0.7071067811865476}, {0.}}, \
{{0.}, {0.}, {0.8660254037844386}}}\
\>", "\<\
0.          -0.866025   0.

0.5         0.707107    0.

0.          0.          0.866025\
\>"], "Output",
 CellChangeTimes->{3.422257525369031*^9, 3.422257659918318*^9, 
  3.422262275051522*^9, 3.422422433839974*^9, 3.422427623357605*^9, 
  3.422433194585268*^9, 3.422513456698519*^9, 3.422513842732745*^9, 
  3.422593230374349*^9, 3.422598797645871*^9, 3.42260192218279*^9, 
  3.422641264272939*^9, 3.422641456310585*^9, 3.422685805683601*^9, 
  3.422686302453876*^9, 3.422696177616648*^9}]
}, Open  ]],

Cell["", "Text",
 CellChangeTimes->{3.42227230000648*^9}],

Cell["Local operators and local basis", "Subtitle",
 CellChangeTimes->{{3.422272304428917*^9, 3.422272311966455*^9}}],

Cell[TextData[{
 "Now let us continue with the construction of the local states. We shall \
represent the operators ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["f", "N"], TraditionalForm]]],
 "by the same operators as before. Then we have four local states: "
}], "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfLocalStates", " ", "=", " ", "2"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"LocalStates", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
    "LocalStates", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], 
    "=", 
    RowBox[{"fDaggerUp", ".", "VacuumState"}]}], ";"}], 
  " "}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "LocalStates", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"fDaggerUp", ".", "fDaggerDown", ".", "VacuumState"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.42226084656336*^9, 3.422260942028706*^9}, {
  3.422261115279298*^9, 3.422261172794231*^9}, {3.422272469310303*^9, 
  3.422272491843473*^9}, {3.422272529795952*^9, 3.422272615717838*^9}, {
  3.422272682500722*^9, 3.422272698956333*^9}, {3.42227335639425*^9, 
  3.422273365754123*^9}, {3.422273443687094*^9, 3.42227344684052*^9}, {
  3.422598805686274*^9, 3.422598813476663*^9}, {3.42260194513799*^9, 
  3.422601950560611*^9}}],

Cell["We check if they are ordered properly:", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderedLocalQuantumNumbers", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfSymmetries", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SzLocal", " ", "=", " ", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"fDaggerUp", ".", "fUp"}], "-", 
      RowBox[{"fDaggerDown", ".", "fDown"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"QzLocal", " ", "=", " ", 
    RowBox[{
     FractionBox["1", "2"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"fDaggerUp", ".", "fUp"}], "+", 
       RowBox[{"fDaggerDown", ".", "fDown"}], " ", "-", " ", "UnitOperator"}],
       ")"}]}]}], ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "QzLocal", ".", 
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "SzLocal", ".", 
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     "\[IndentingNewLine]", 
     RowBox[{"{", " ", 
      RowBox[{
       RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", " ", 
       RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], " ", "//",
   " ", "MatrixForm"}]}], "Input",
 CellChangeTimes->{{3.422261008583634*^9, 3.422261105170266*^9}, {
   3.422261185877381*^9, 3.422261257246138*^9}, {3.422261290925316*^9, 
   3.42226138106269*^9}, {3.422261556601073*^9, 3.422261621932931*^9}, {
   3.422261667725704*^9, 3.422261669214305*^9}, {3.422261754271065*^9, 
   3.422261761177845*^9}, {3.422261827807973*^9, 3.422261982154367*^9}, {
   3.422272998493306*^9, 3.422273023589198*^9}, {3.422273060650893*^9, 
   3.42227306270624*^9}, {3.422273102646266*^9, 3.422273104921433*^9}, {
   3.422273156211252*^9, 3.422273179056414*^9}, {3.42227334592718*^9, 
   3.422273346665373*^9}, 3.422427646446775*^9, {3.422427701122459*^9, 
   3.422427780430807*^9}, {3.42242781943902*^9, 3.422427835662201*^9}, 
   3.422593238762993*^9, {3.422601965810531*^9, 3.422601985920155*^9}}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", 
      FractionBox["1", "2"]},
     {
      FractionBox["1", "2"], "0"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{
  3.422273025881706*^9, {3.422273077127199*^9, 3.422273105554888*^9}, 
   3.422273179924059*^9, {3.422273347766973*^9, 3.422273369596677*^9}, 
   3.422273452148074*^9, 3.42242243891415*^9, {3.422427783820186*^9, 
   3.422427793470396*^9}, 3.422427836351067*^9, 3.422433199232154*^9, 
   3.422513473454277*^9, 3.422513846525845*^9, {3.422593237016459*^9, 
   3.422593241662812*^9}, 3.422598824354917*^9, {3.422601980039036*^9, 
   3.42260198658675*^9}, 3.422641467306709*^9, 3.422685809889307*^9, 
   3.422696186542167*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"OrderedLocalfDaggerUp", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"OrderedLocalfDaggerDown", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"OrderedLocalfUp", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedLocalfDown", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]", "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"OrderedLocalfDaggerUp", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDaggerUp", ".", 
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedLocalfDaggerDown", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDaggerDown", ".", 
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedLocalfUp", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fUp", ".", 
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedLocalfDown", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDown", ".", 
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"j", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SpinReducedLocalfDagger", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SpinReducedLocalf", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ReducedLocalf", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]", "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Table", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"OrderedLocalfDaggerUp", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]",
            "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SpinReducedLocalfDagger", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ",
            
           RowBox[{
            RowBox[{"OrderedLocalfDaggerUp", "\[LeftDoubleBracket]", 
             RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
            RowBox[{"ClebschGordan", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                FractionBox["1", "2"], ",", 
                FractionBox["1", "2"]}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], 
               "}"}]}], "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"OrderedLocalfDaggerDown", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]",
            "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SpinReducedLocalfDagger", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ",
            
           RowBox[{
            RowBox[{"OrderedLocalfDaggerDown", "\[LeftDoubleBracket]", 
             RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
            RowBox[{"ClebschGordan", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                FractionBox["1", "2"], ",", 
                RowBox[{"-", 
                 FractionBox["1", "2"]}]}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], 
               "}"}]}], "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        "\[IndentingNewLine]", 
        RowBox[{"{", 
         RowBox[{"SpinReducedLocalfDagger", "\[LeftDoubleBracket]", 
          RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
       " ", 
       RowBox[{"{", 
        RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"j", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], "//", 
     "MatrixForm"}], "//", "OutputForm"}], "//", "N"}], "\[IndentingNewLine]",
   "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Table", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"OrderedLocalfUp", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]",
            "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SpinReducedLocalf", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ",
            
           RowBox[{
            RowBox[{"OrderedLocalfUp", "\[LeftDoubleBracket]", 
             RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
            RowBox[{"ClebschGordan", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                FractionBox["1", "2"], ",", 
                RowBox[{"-", 
                 FractionBox["1", "2"]}]}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], 
               "}"}]}], "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"OrderedLocalfDown", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]",
            "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SpinReducedLocalf", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ",
            
           RowBox[{
            RowBox[{"-", 
             RowBox[{"OrderedLocalfDown", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}]}], "/", 
            RowBox[{"ClebschGordan", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
              ",", 
              RowBox[{"{", 
               RowBox[{
                FractionBox["1", "2"], ",", 
                FractionBox["1", "2"]}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
                RowBox[{
                "OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                 RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], 
               "}"}]}], "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        "\[IndentingNewLine]", 
        RowBox[{"{", 
         RowBox[{"SpinReducedLocalf", "\[LeftDoubleBracket]", 
          RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
       " ", 
       RowBox[{"{", 
        RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"j", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], "//", 
     "MatrixForm"}], "//", "OutputForm"}], "//", "N"}], "\[IndentingNewLine]",
   "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"SpinReducedLocalfDagger", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedLocalf", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"SpinReducedLocalfDagger", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               FractionBox["1", "2"]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"SpinReducedLocalf", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedLocalf", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"SpinReducedLocalf", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               RowBox[{"-", 
                FractionBox["1", "2"]}]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"ReducedLocalf", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
      " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422262166469301*^9, 3.422262185711958*^9}, {
   3.422262232176988*^9, 3.422262232892694*^9}, {3.42227287163859*^9, 
   3.422272880981803*^9}, {3.4224229432499*^9, 3.422422953983789*^9}, {
   3.422427878701063*^9, 3.422427932721799*^9}, {3.422433118922209*^9, 
   3.42243312210934*^9}, 3.422593243403996*^9, 3.422598846334679*^9, {
   3.422602056138283*^9, 3.422602389105671*^9}, 3.422602440521864*^9, {
   3.422602626528262*^9, 3.422602660702424*^9}, {3.422602776895065*^9, 
   3.422602809698303*^9}, {3.422641510692696*^9, 3.422641527798756*^9}, {
   3.42268591903634*^9, 3.422685922314098*^9}, {3.42268617215136*^9, 
   3.42268617437606*^9}}],

Cell[OutputFormData["\<\
{{{0.}, {0.}}, {{-1.4142135623730951}, {0.}}}\
\>", "\<\
0.         0.

-1.41421   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422257511474477*^9, 3.422262186465329*^9, 3.422262234138565*^9, 
   3.422272883289382*^9, 3.422273385171968*^9, 3.422273462344101*^9, 
   3.422422442190021*^9, {3.422422946203317*^9, 3.422422954776595*^9}, 
   3.422427934223118*^9, 3.422433201722852*^9, 3.422513499654047*^9, 
   3.422513847815914*^9, 3.422593244149783*^9, 3.422598858699624*^9, 
   3.422602000468955*^9, {3.42260236963319*^9, 3.422602391146996*^9}, 
   3.422602441299971*^9, {3.422602631654118*^9, 3.42260266151668*^9}, {
   3.422602771377556*^9, 3.422602811191511*^9}, {3.422641496322143*^9, 
   3.422641528648566*^9}, 3.422685813461894*^9, 3.422685923310854*^9, {
   3.422686175656952*^9, 3.422686203978569*^9}, 3.42269618780487*^9}],

Cell[OutputFormData["\<\
{{{0.}, {1.}}, {{0.}, {0.}}}\
\>", "\<\
0.   1.

0.   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422257511474477*^9, 3.422262186465329*^9, 3.422262234138565*^9, 
   3.422272883289382*^9, 3.422273385171968*^9, 3.422273462344101*^9, 
   3.422422442190021*^9, {3.422422946203317*^9, 3.422422954776595*^9}, 
   3.422427934223118*^9, 3.422433201722852*^9, 3.422513499654047*^9, 
   3.422513847815914*^9, 3.422593244149783*^9, 3.422598858699624*^9, 
   3.422602000468955*^9, {3.42260236963319*^9, 3.422602391146996*^9}, 
   3.422602441299971*^9, {3.422602631654118*^9, 3.42260266151668*^9}, {
   3.422602771377556*^9, 3.422602811191511*^9}, {3.422641496322143*^9, 
   3.422641528648566*^9}, 3.422685813461894*^9, 3.422685923310854*^9, {
   3.422686175656952*^9, 3.422686203978569*^9}, 3.422696187833586*^9}],

Cell[OutputFormData["\<\
{{{0.}, {1.4142135623730951}}, {{-1.4142135623730951}, {0.}}}\
\>", "\<\
0.         1.41421

-1.41421   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422257511474477*^9, 3.422262186465329*^9, 3.422262234138565*^9, 
   3.422272883289382*^9, 3.422273385171968*^9, 3.422273462344101*^9, 
   3.422422442190021*^9, {3.422422946203317*^9, 3.422422954776595*^9}, 
   3.422427934223118*^9, 3.422433201722852*^9, 3.422513499654047*^9, 
   3.422513847815914*^9, 3.422593244149783*^9, 3.422598858699624*^9, 
   3.422602000468955*^9, {3.42260236963319*^9, 3.422602391146996*^9}, 
   3.422602441299971*^9, {3.422602631654118*^9, 3.42260266151668*^9}, {
   3.422602771377556*^9, 3.422602811191511*^9}, {3.422641496322143*^9, 
   3.422641528648566*^9}, 3.422685813461894*^9, 3.422685923310854*^9, {
   3.422686175656952*^9, 3.422686203978569*^9}, 3.422696187866855*^9}]
}, Open  ]],

Cell["\<\
We write the quantum numbers to file \"QuantumNumbers\" (spin quantum numbers \
are multiplied by 2):\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428522554788*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./QuantumNumbers\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"WriteString", "[", " ", 
      RowBox[{"strm", ",", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
          RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], " ", "]"}], 
        "]"}], ",", " ", "\"\<\\t\>\"", ",", " ", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"2", " ", 
          RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
           RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "]"}], "]"}],
        ",", " ", "\"\<\\n\>\""}], "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfBlockStates"}], "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422428031091177*^9, 3.422428228324012*^9}, 
   3.422428274771525*^9, {3.422428334035563*^9, 3.422428414302918*^9}, {
   3.422433090180519*^9, 3.422433103257273*^9}}],

Cell["\<\
Next, the hopping, spectral and Hamiltonian operators are saved in \
\"HoppingOperators\", \"SpectralOperators\", and \"HamiltonianBlocks\", \
respectively:\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428580988006*^9}, {3.422428618117171*^9, 3.422428703986899*^9}, {
  3.422513558824415*^9, 3.422513564019709*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./HoppingOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<ReducedfDagger\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"Reducedf", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{
  3.42251352727455*^9, {3.422513570627052*^9, 3.422513657546017*^9}, 
   3.422598917156166*^9, {3.422641563776373*^9, 3.422641564699104*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./SpectralOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<Spin\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"ReducedS", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422513664909368*^9, 3.422513717313067*^9}, {
  3.42264158790005*^9, 3.422641594993684*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./HamiltonianBlocks\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"For", "[", 
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", 
    RowBox[{"k", "\[LessEqual]", " ", "NumberOfHamiltonianTerms"}], ",", " ", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Write", "[", 
      RowBox[{
      "strm", ",", "\"\<HamiltonianBlock \>\"", ",", "k", ",", 
       "\"\<\\n\>\""}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"For", "[", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", 
       RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"For", "[", 
         RowBox[{
          RowBox[{"j", "=", "1"}], ",", 
          RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", 
             RowBox[{"N", "[", 
              RowBox[{"CForm", "[", 
               RowBox[{"OrderedHamiltonian", "\[LeftDoubleBracket]", 
                RowBox[{"k", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], 
               " ", "]"}], "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
             RowBox[{"WriteString", "[", 
              RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"WriteString", "[", 
         RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"i", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"Write", "[", 
      RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", "\[IndentingNewLine]",
      
     RowBox[{"k", "++"}]}]}], "]"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422422461984407*^9, 3.422422590368324*^9}, {
  3.422422657679912*^9, 3.422422675770448*^9}, {3.422422769910713*^9, 
  3.422422774310689*^9}, {3.422422907355884*^9, 3.422422907479271*^9}, {
  3.422423014025691*^9, 3.422423024399214*^9}, {3.422423065573881*^9, 
  3.422423075136177*^9}, {3.422428735589752*^9, 3.422428797092516*^9}, {
  3.42242951569936*^9, 3.422429705585544*^9}, {3.422429793487009*^9, 
  3.4224298166685*^9}, {3.422429927802331*^9, 3.422429930748695*^9}, {
  3.422432038815338*^9, 3.422432247503146*^9}, {3.422432279060799*^9, 
  3.422432423000301*^9}, {3.422432470625659*^9, 3.422432486448996*^9}, {
  3.422513762990068*^9, 3.422513804257083*^9}}],

Cell["\<\
Finally, we write the local quantum numbers to and local operators to file \
\"LocalQuantumNumbers\" and \"LocalHoppingOperators\" (spin quantum numbers \
are multiplied by 2):\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428522554788*^9}, {3.422432507395505*^9, 3.422432527004284*^9}, {
  3.422432571670566*^9, 3.422432576538502*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./LocalQuantumNumbers\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"WriteString", "[", " ", 
      RowBox[{"strm", ",", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
          RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], " ", "]"}], 
        "]"}], ",", " ", "\"\<\\t\>\"", ",", " ", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"2", " ", 
          RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
           RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "]"}], "]"}],
        ",", " ", "\"\<\\n\>\""}], "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfLocalStates"}], "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Close", "[", "strm", "]"}], ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./LocalHoppingOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<ReducedLocalfDagger\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfLocalStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfLocalStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"ReducedLocalf", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfLocalStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422432891090526*^9, 3.422433072089265*^9}, 
   3.42243322189554*^9, 3.42259893574203*^9, {3.422641607864633*^9, 
   3.422641608767049*^9}}]
},
WindowSize->{1093, 645},
WindowMargins->{{37, Automatic}, {9, Automatic}},
Magnification->1,
FrontEndVersion->"6.0 for Linux x86 (32-bit) (April 20, 2007)",
StyleDefinitions->"Classroom.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[568, 21, 437, 16, 60, "Title"],
Cell[1008, 39, 35, 0, 40, "Subtitle"],
Cell[1046, 41, 234, 5, 63, "Input"],
Cell[1283, 48, 3955, 119, 273, "Text"],
Cell[5241, 169, 361, 8, 64, "Input"],
Cell[5605, 179, 331, 6, 24, "Text"],
Cell[5939, 187, 4167, 102, 429, "Input"],
Cell[10109, 291, 758, 12, 24, "Text"],
Cell[10870, 305, 4337, 130, 598, "Input"],
Cell[15210, 437, 119, 1, 24, "Text"],
Cell[15332, 440, 26, 0, 40, "Subtitle"],
Cell[15361, 442, 554, 9, 58, "Text"],
Cell[15918, 453, 303, 8, 69, "Input"],
Cell[16224, 463, 1843, 50, 158, "Input"],
Cell[18070, 515, 234, 5, 24, "Text"],
Cell[CellGroupData[{
Cell[18329, 524, 2719, 57, 183, "Input"],
Cell[21051, 583, 1842, 37, 107, "Output"]
}, Open  ]],
Cell[22908, 623, 182, 4, 24, "Text"],
Cell[23093, 629, 51, 0, 40, "Subtitle"],
Cell[CellGroupData[{
Cell[23169, 633, 16303, 368, 1220, "Input"],
Cell[39475, 1003, 1288, 25, 103, "Output"],
Cell[40766, 1030, 1228, 24, 103, "Output"],
Cell[41997, 1056, 1334, 25, 103, "Output"]
}, Open  ]],
Cell[43346, 1084, 37, 0, 40, "Subtitle"],
Cell[43386, 1086, 698, 19, 42, "Text"],
Cell[CellGroupData[{
Cell[44109, 1109, 2647, 68, 201, "Input"],
Cell[46759, 1179, 572, 13, 67, "Output"]
}, Open  ]],
Cell[47346, 1195, 99, 1, 40, "Subtitle"],
Cell[47448, 1198, 176, 2, 24, "Text"],
Cell[CellGroupData[{
Cell[47649, 1204, 8284, 194, 533, "Input"],
Cell[55936, 1400, 640, 15, 103, "Output"]
}, Open  ]],
Cell[56591, 1418, 57, 1, 24, "Text"],
Cell[56651, 1421, 117, 1, 40, "Subtitle"],
Cell[56771, 1424, 387, 9, 24, "Text"],
Cell[57161, 1435, 1231, 32, 115, "Input"],
Cell[58395, 1469, 263, 3, 24, "Text"],
Cell[CellGroupData[{
Cell[58683, 1476, 3112, 72, 239, "Input"],
Cell[61798, 1550, 1176, 28, 97, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[63011, 1583, 16660, 382, 1328, "Input"],
Cell[79674, 1967, 855, 17, 85, "Output"],
Cell[80532, 1986, 827, 17, 85, "Output"],
Cell[81362, 2005, 877, 17, 85, "Output"]
}, Open  ]],
Cell[82254, 2025, 432, 8, 24, "Text"],
Cell[82689, 2035, 1484, 34, 131, "Input"],
Cell[84176, 2071, 582, 10, 24, "Text"],
Cell[84761, 2083, 2089, 52, 250, "Input"],
Cell[86853, 2137, 2032, 51, 250, "Input"],
Cell[88888, 2190, 3047, 67, 284, "Input"],
Cell[91938, 2259, 602, 10, 24, "Text"],
Cell[92543, 2271, 3526, 86, 438, "Input"]
}
]
*)

(* End of internal cache information *)
